/*******************************************************************************

NAME            nvme-core.c
UPDATE DATE     %date_modified: Thu Jun 08 10:09:07 2017 %
PROGRAMMER      %created_by:    mben %

    Copyright 2017-2018 NetApp, Inc.

DESCRIPTION:
    An kernel driver to support a user-space implementation of an NVMe "device".

Change Log (for open source compliance):
    (2017-06-09):Replaced PCI device items with new user space controller objects.
    (2017-06-14):Replaced nvme_dev struct with nub_dev struct
    (2017-06-15):Switched to using a controller memory buffer for internal command buffers.
    (2017-06-19):Used controller memory buffer for PRP list.
    (2017-06-23):Added several IOCTL calls necessary for proper initialization and run time communication with the user space driver.
    (2017-07-11):Added debug mechanisms found useful during development.
    (2017-07-21):Fixed issues with the data copy IOCTL.
    (2017-08-04):Removed unnecessary inclusion of nvme.h and unused IOCTLs.
    (2017-08-10):Reworked block reads to address some issues.
    (2017-08-15):Added workqueues to handle completions.
    (2017-08-28):Cleaned up some issues with the shutdown code.
    (2017-09-20):Fixed some IO issues discovered.
    (2017-10-01):Fixed some issues with Large IO and IOs on certain boundaries.
    (2017-10-05):Added additional trace logging for debugging issues.
    (2017-12-11):Added ability to change size of, remove or add a block device at runtime.
    (2018-01-09):Added IOCTL to allow for block device updates.
    (2018-01-12):Changed to handle completions in the client context for improved performance.
    (2018-02-16):Added PI checking to the client data copy ioctl for improved performance.
    (2018-03-05):Increased max IO size to 2MB.
    (2018-03-26):Fixed issue with 8K IO sizes on certain boundaries.
    (2018-04-23):Added changlog for compliance per legal guidance.
    (2018-08-02):Varied code to compile out all content if compiling for kernel level
                 greater than 4.9.0

*******************************************************************************/

/*
 * NVM Express device driver
 * Copyright (c) 2011-2014, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 */

/* Until the NUBD driver is updated for stretch/buster kernel, we will
 * have to disable this driver by not compiling in much of the source
 */
#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,9,0)
    #define DISABLE_NUBD_DRIVER
#endif

#include <linux/bio.h>
#include <linux/bitops.h>
#include <linux/blkdev.h>
#include <linux/cpu.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/genhd.h>
#include <linux/hdreg.h>
#include <linux/idr.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/kdev_t.h>
#include <linux/kthread.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include "linux/nubd.h"
#include <linux/pci.h>
#include <linux/percpu.h>
#include <linux/poison.h>
#include <linux/ptrace.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <scsi/sg.h>
#ifndef DISABLE_NUBD_DRIVER
    #include <asm-generic/io-64-nonatomic-lo-hi.h>
    #include <uapi/linux/nvme.h>
#endif

#include <trace/events/block.h>

#ifndef DISABLE_NUBD_DRIVER

#define DEFAULT_ADMIN_QDEPTH    64

#define NVME_Q_DEPTH        1024
#define SQ_SIZE(depth)      (depth * sizeof(struct nvme_command))
#define CQ_SIZE(depth)      (depth * sizeof(struct nvme_completion))
#define ADMIN_TIMEOUT       (admin_timeout * HZ)
#define IOD_TIMEOUT     (retry_time * HZ)

/* This macro translates a dma address in the CMB to a kernel virtual address */
#define KV_CMB_ADDR(dev, addr)  (void*) (dev->cmb_vbase + ((uintptr_t)addr - (uintptr_t)dev->cmb_pbase))

#define NUB_KDEV(ndev)  ndev->miscdev.this_device

#define BYTES_LEFT_IN_PAGE(x) (PAGE_SIZE - ((size_t)(x) & (PAGE_SIZE-1)))

/*
 * Number of seconds to allow the namespace modification
 * ioctls to wait before returning error.
 */
#define NAMESPACE_LOCK_TIMEOUT  1

/*
 * The application should be polling, so technically writing the doorbell
 * value should not be needed.  We do it here to stay congruent to the hardware
 * implementation, and as it may provide some debug information if things
 * go wrong.
 */
#define SQ_DB_NOTIFY(tail, db_addr)     writel(tail, db_addr)

static unsigned char admin_timeout = 60;
module_param(admin_timeout, byte, 0644);
MODULE_PARM_DESC(admin_timeout, "timeout in seconds for admin commands");

unsigned char nvme_io_timeout = 60;
module_param_named(io_timeout, nvme_io_timeout, byte, 0644);
MODULE_PARM_DESC(io_timeout, "timeout in seconds for I/O");

static unsigned char retry_time = 60;
module_param(retry_time, byte, 0644);
MODULE_PARM_DESC(retry_time, "time in seconds to retry failed I/O");

/*
 * We should be able to support up to 26, but keep this as a smaller value
 * as we really only expect to have one controller for the foreseeable future.
 */
static unsigned char max_ctrls = 4;
module_param(max_ctrls, byte, 0644);
MODULE_PARM_DESC(max_ctrls, "Maximum number of nubd controller devices");

static int nubd_major;
module_param(nubd_major, int, 0444);
MODULE_PARM_DESC(nubd_major, "Major number to be used for block devices");

static int clnt_completion = 1;
module_param(clnt_completion, int, 0644);
MODULE_PARM_DESC(clnt_completion, "Completions handled in client context (performance tunable)");

static DEFINE_IDA(nubd_ctrl_ida);
static DEFINE_SEMAPHORE(ns_update_lock);

static struct miscdevice nubd_mgr;

/* Tracking array for nub controller device instances. */
static struct nub_dev** nubd_ctrls;

static DEFINE_SPINLOCK(dev_list_lock);
static LIST_HEAD(dev_list);
static struct task_struct *nvme_thread;
static struct workqueue_struct *nvme_workq;
static wait_queue_head_t nvme_kthread_wait;

static void nubd_remove(struct nub_dev *dev);
static void nvme_reset_failed_dev(struct work_struct *ws);
static void* nubd_alloc_prp_list(struct nub_dev* dev, dma_addr_t *handle);
static void nubd_free_prp_list(struct nub_dev* dev, void* vaddr);
static int  nvme_ns_update(struct nub_dev *dev);
static int  nubd_update_disk(struct nub_dev *dev, int nsid);

extern u16 crc16_t10dif_copy_by4(
               u16 init_crc,
               unsigned char *dst,
               const unsigned char *src,
               u64 len
       );

/* We only support type 1 DIF information. */
struct t10_dif_tuple {
    __be16 guard_tag;
    __be16 app_tag;     /* Fixed value */
    __be32 ref_tag;     /* Lower 32 bits of LBA */
};

struct async_cmd_info {
    struct kthread_work work;
    struct kthread_worker *worker;
    u32 result;
    int status;
    void *ctx;
};

/*
 * An NVM Express queue.  Each device has at least two (one for admin
 * commands and one for I/O commands).
 */
struct nvme_queue {
    struct rcu_head r_head;
    struct device *q_dmadev;
    struct nub_dev *dev;
    struct work_struct cq_update_work;

    struct workqueue_struct *io_workqueue;
    char irqname[24];   /* nvme4294967295-65535\0 */
    spinlock_t q_lock;
    struct nvme_command *sq_cmds;
    volatile struct nvme_completion *cqes;
    dma_addr_t sq_dma_addr;
    dma_addr_t cq_dma_addr;
    wait_queue_head_t sq_full;
    wait_queue_t sq_cong_wait;
    struct bio_list sq_cong;
    struct list_head iod_bio;
    u32 __iomem *q_db;
    u16 q_depth;
    u16 sq_head;
    u16 sq_tail;
    u16 cq_head;
    u16 qid;
    u8 cq_phase;
    u8 cqe_seen;
    u8 q_suspended;
    cpumask_var_t cpu_mask;
    struct async_cmd_info cmdinfo;
    unsigned long cmdid_data[];
};

/*
 * Check we didin't inadvertently grow the command struct
 */
static inline void _nvme_check_size(void)
{
    BUILD_BUG_ON(sizeof(struct nvme_rw_command) != 64);
    BUILD_BUG_ON(sizeof(struct nvme_create_cq) != 64);
    BUILD_BUG_ON(sizeof(struct nvme_create_sq) != 64);
    BUILD_BUG_ON(sizeof(struct nvme_delete_queue) != 64);
    BUILD_BUG_ON(sizeof(struct nvme_features) != 64);
    BUILD_BUG_ON(sizeof(struct nvme_format_cmd) != 64);
    BUILD_BUG_ON(sizeof(struct nvme_abort_cmd) != 64);
    BUILD_BUG_ON(sizeof(struct nvme_command) != 64);
    BUILD_BUG_ON(sizeof(struct nvme_id_ctrl) != 4096);
    BUILD_BUG_ON(sizeof(struct nvme_id_ns) != 4096);
    BUILD_BUG_ON(sizeof(struct nvme_lba_range_type) != 64);
    BUILD_BUG_ON(sizeof(struct nvme_smart_log) != 512);
}

typedef void (*nvme_completion_fn)(struct nvme_queue *, void *,
                        struct nvme_completion *);

struct nvme_cmd_info {
    nvme_completion_fn fn;
    void *ctx;
    unsigned long timeout;
    int aborted;
};

struct io_region_entry {
    struct  io_region_entry* next;
    void*   v_addr;
    u64     p_addr;
    size_t  len;
};

struct io_region_entry* io_region_list = NULL;

static struct nvme_cmd_info *nvme_cmd_info(struct nvme_queue *nvmeq)
{
    return (void *)&nvmeq->cmdid_data[BITS_TO_LONGS(nvmeq->q_depth)];
}

static unsigned nvme_queue_extra(int depth)
{
    return DIV_ROUND_UP(depth, 8) + (depth * sizeof(struct nvme_cmd_info));
}

/**
 * alloc_cmdid() - Allocate a Command ID
 * @nvmeq: The queue that will be used for this command
 * @ctx: A pointer that will be passed to the handler
 * @handler: The function to call on completion
 *
 * Allocate a Command ID for a queue.  The data passed in will
 * be passed to the completion handler.  This is implemented by using
 * the bottom two bits of the ctx pointer to store the handler ID.
 * Passing in a pointer that's not 4-byte aligned will cause a BUG.
 * We can change this if it becomes a problem.
 *
 * May be called with local interrupts disabled and the q_lock held,
 * or with interrupts enabled and no locks held.
 */
static int alloc_cmdid(struct nvme_queue *nvmeq, void *ctx,
                nvme_completion_fn handler, unsigned timeout)
{
    int depth = nvmeq->q_depth - 1;
    struct nvme_cmd_info *info = nvme_cmd_info(nvmeq);
    int cmdid;

    do {
        cmdid = find_first_zero_bit(nvmeq->cmdid_data, depth);
        if (cmdid >= depth)
            return -EBUSY;
    } while (test_and_set_bit(cmdid, nvmeq->cmdid_data));

    info[cmdid].fn = handler;
    info[cmdid].ctx = ctx;
    info[cmdid].timeout = jiffies + timeout;
    info[cmdid].aborted = 0;
    return cmdid;
}

static int alloc_cmdid_killable(struct nvme_queue *nvmeq, void *ctx,
                nvme_completion_fn handler, unsigned timeout)
{
    int cmdid;
    wait_event_killable(nvmeq->sq_full,
        (cmdid = alloc_cmdid(nvmeq, ctx, handler, timeout)) >= 0);
    return (cmdid < 0) ? -EINTR : cmdid;
}

/* Special values must be less than 0x1000 */
#define CMD_CTX_BASE        ((void *)POISON_POINTER_DELTA)
#define CMD_CTX_CANCELLED   (0x30C + CMD_CTX_BASE)
#define CMD_CTX_COMPLETED   (0x310 + CMD_CTX_BASE)
#define CMD_CTX_INVALID     (0x314 + CMD_CTX_BASE)
#define CMD_CTX_ABORT       (0x318 + CMD_CTX_BASE)

static void special_completion(struct nvme_queue *nvmeq, void *ctx,
                        struct nvme_completion *cqe)
{
    if (ctx == CMD_CTX_CANCELLED)
        return;
    if (ctx == CMD_CTX_ABORT) {
        ++nvmeq->dev->abort_limit;
        return;
    }
    if (ctx == CMD_CTX_COMPLETED) {
        dev_warn(nvmeq->q_dmadev,
                "completed id %d twice on queue %d\n",
                cqe->command_id, le16_to_cpup(&cqe->sq_id));
        return;
    }
    if (ctx == CMD_CTX_INVALID) {
        dev_warn(nvmeq->q_dmadev,
                "invalid id %d completed on queue %d\n",
                cqe->command_id, le16_to_cpup(&cqe->sq_id));
        return;
    }

    dev_warn(nvmeq->q_dmadev, "Unknown special completion %p\n", ctx);
}

static void async_completion(struct nvme_queue *nvmeq, void *ctx,
                        struct nvme_completion *cqe)
{
    struct async_cmd_info *cmdinfo = ctx;
    cmdinfo->result = le32_to_cpup(&cqe->result);
    cmdinfo->status = le16_to_cpup(&cqe->status) >> 1;
    queue_kthread_work(cmdinfo->worker, &cmdinfo->work);
}

/*
 * Called with local interrupts disabled and the q_lock held.  May not sleep.
 */
static void *free_cmdid(struct nvme_queue *nvmeq, int cmdid,
                        nvme_completion_fn *fn)
{
    void *ctx;
    struct nvme_cmd_info *info = nvme_cmd_info(nvmeq);

    if (cmdid >= nvmeq->q_depth || !info[cmdid].fn) {
        if (fn)
            *fn = special_completion;
        return CMD_CTX_INVALID;
    }
    if (fn)
        *fn = info[cmdid].fn;
    ctx = info[cmdid].ctx;
    info[cmdid].fn = special_completion;
    info[cmdid].ctx = CMD_CTX_COMPLETED;
    clear_bit(cmdid, nvmeq->cmdid_data);
    wake_up(&nvmeq->sq_full);
    return ctx;
}

static void *cancel_cmdid(struct nvme_queue *nvmeq, int cmdid,
                        nvme_completion_fn *fn)
{
    void *ctx;
    struct nvme_cmd_info *info = nvme_cmd_info(nvmeq);
    if (fn)
        *fn = info[cmdid].fn;
    ctx = info[cmdid].ctx;
    info[cmdid].fn = special_completion;
    info[cmdid].ctx = CMD_CTX_CANCELLED;
    return ctx;
}

static struct nvme_queue *raw_nvmeq(struct nub_dev *dev, int qid)
{
    return rcu_dereference_raw(dev->queues[qid]);
}

static struct nvme_queue *get_nvmeq(struct nub_dev *dev) __acquires(RCU)
{
    struct nvme_queue *nvmeq;
    unsigned queue_id = get_cpu_var(*dev->io_queue);

    rcu_read_lock();
    nvmeq = rcu_dereference(dev->queues[queue_id]);
    if (nvmeq)
        return nvmeq;

    rcu_read_unlock();
    put_cpu_var(*dev->io_queue);
    return NULL;
}

static void put_nvmeq(struct nvme_queue *nvmeq) __releases(RCU)
{
    rcu_read_unlock();
    put_cpu_var(nvmeq->dev->io_queue);
}

static struct nvme_queue *lock_nvmeq(struct nub_dev *dev, int q_idx)
                            __acquires(RCU)
{
    struct nvme_queue *nvmeq;

    rcu_read_lock();
    nvmeq = rcu_dereference(dev->queues[q_idx]);
    if (nvmeq)
        return nvmeq;

    rcu_read_unlock();
    return NULL;
}

static void unlock_nvmeq(struct nvme_queue *nvmeq) __releases(RCU)
{
    rcu_read_unlock();
}

/**
 * nvme_submit_cmd() - Copy a command into a queue and ring the doorbell
 * @nvmeq: The queue to use
 * @cmd: The command to send
 *
 * Safe to use from interrupt context
 */
static int nvme_submit_cmd(struct nvme_queue *nvmeq, struct nvme_command *cmd)
{
    unsigned long flags;
    u16 tail;
    spin_lock_irqsave(&nvmeq->q_lock, flags);
    if (nvmeq->q_suspended) {
        spin_unlock_irqrestore(&nvmeq->q_lock, flags);
        return -EBUSY;
    }
    tail = nvmeq->sq_tail;
    memcpy(&nvmeq->sq_cmds[tail], cmd, sizeof(*cmd));
    if (++tail == nvmeq->q_depth)
        tail = 0;
    SQ_DB_NOTIFY(tail, nvmeq->q_db);
    nvmeq->sq_tail = tail;
    spin_unlock_irqrestore(&nvmeq->q_lock, flags);

    return 0;
}

static __le64 **iod_list(struct nvme_iod *iod)
{
    return ((void *)iod) + iod->offset;
}

/*
 * Will slightly overestimate the number of pages needed.  This is OK
 * as it only leads to a small amount of wasted memory for the lifetime of
 * the I/O.
 */
static int nvme_npages(unsigned size)
{
    unsigned nprps = DIV_ROUND_UP(size + PAGE_SIZE, PAGE_SIZE);
    return DIV_ROUND_UP(8 * nprps, PAGE_SIZE - 8);
}

static struct nvme_iod *
nvme_alloc_iod(unsigned nseg, unsigned nbytes, gfp_t gfp)
{
    struct nvme_iod *iod = kmalloc(sizeof(struct nvme_iod) +
                sizeof(__le64 *) * nvme_npages(nbytes) +
                sizeof(struct scatterlist) * nseg, gfp);

    if (iod) {
        iod->offset = offsetof(struct nvme_iod, sg[nseg]);
        iod->npages = -1;
        iod->length = nbytes;
        iod->nents = 0;
        iod->first_dma = 0ULL;
        iod->start_time = jiffies;
    }

    return iod;
}


/*
 * These functions to allocate and free command buffers
 * are implemented separately though with minimal functionality
 * so we can easily swap out the implementation for different
 * situations as needed (e.g, PCI device vs user space implementation).
 */
void* nvme_alloc_cmd_buffer(struct nub_dev* dev,
            size_t len, dma_addr_t* p_addr)
{
    void* mem;

    mem = kmalloc(len, GFP_KERNEL);
    if (mem)
    {
        *p_addr = virt_to_phys(mem);
    }

    return mem;
}

void  nvme_free_cmd_buffer(struct nub_dev* dev, size_t len,
            void* v_addr, dma_addr_t p_addr)
{
    kfree(v_addr);
}

void nvme_free_iod(struct nub_dev *dev, struct nvme_iod *iod)
{
    int i;
    __le64 **list = iod_list(iod);

    if (iod->npages == 0)
        nubd_free_prp_list(dev, list[0]);

    for (i = 0; i < iod->npages; i++) {
        __le64 *prp_list = list[i];
        nubd_free_prp_list(dev, prp_list);
    }
    kfree(iod);
}

static void nvme_start_io_acct(struct bio *bio)
{
    struct gendisk *disk = bio->bi_bdev->bd_disk;
    if (blk_queue_io_stat(disk->queue)) {
        const int rw = bio_data_dir(bio);
        int cpu = part_stat_lock();
        part_round_stats(cpu, &disk->part0);
        part_stat_inc(cpu, &disk->part0, ios[rw]);
        part_stat_add(cpu, &disk->part0, sectors[rw],
                            bio_sectors(bio));
        part_inc_in_flight(&disk->part0, rw);
        part_stat_unlock();
    }
}

static void nvme_end_io_acct(struct bio *bio, unsigned long start_time)
{
    struct gendisk *disk = bio->bi_bdev->bd_disk;
    if (blk_queue_io_stat(disk->queue)) {
        const int rw = bio_data_dir(bio);
        unsigned long duration = jiffies - start_time;
        int cpu = part_stat_lock();
        part_stat_add(cpu, &disk->part0, ticks[rw], duration);
        part_round_stats(cpu, &disk->part0);
        part_dec_in_flight(&disk->part0, rw);
        part_stat_unlock();
    }
}

static void bio_completion(struct nvme_queue *nvmeq, void *ctx,
                        struct nvme_completion *cqe)
{
    struct nvme_iod *iod = ctx;
    struct bio *bio = iod->private;
    u16 status = le16_to_cpup(&cqe->status) >> 1;
    int error = 0;

    if (unlikely(status)) {
        dev_warn( NUB_KDEV(nvmeq->dev), "%s: Bad status (0x%x) for cmdId (0x%x), sqId (0x%x).",
                 __FUNCTION__, status, cqe->command_id, cqe->sq_id);
        if (!(status & NVME_SC_DNR ||
                bio->bi_rw & REQ_FAILFAST_MASK) &&
                (jiffies - iod->start_time) < IOD_TIMEOUT) {
            if (!waitqueue_active(&nvmeq->sq_full))
                add_wait_queue(&nvmeq->sq_full,
                            &nvmeq->sq_cong_wait);
            list_add_tail(&iod->node, &nvmeq->iod_bio);
            wake_up(&nvmeq->sq_full);
            return;
        }
        error = -EIO;
    }
    if (iod->nents) {
    /*
     * TODO US167552: If we need to allow EOS DMA HW access, we probably need
     * to map this in here.  Remove for time being though, doing CPU copy.
     */
#if 0
        dma_unmap_sg(nvmeq->q_dmadev, iod->sg, iod->nents,
            bio_data_dir(bio) ? DMA_TO_DEVICE : DMA_FROM_DEVICE);
#endif
        nvme_end_io_acct(bio, iod->start_time);
    }
    nvme_free_iod(nvmeq->dev, iod);

    trace_block_bio_complete(bdev_get_queue(bio->bi_bdev), bio, error);
    bio_endio(bio, error);
}

/* length is in bytes.  gfp flags indicates whether we may sleep. */
int nvme_setup_prps(struct nub_dev *dev, struct nvme_iod *iod, int total_len,
                                gfp_t gfp)
{
    int length = total_len;
    struct scatterlist *sg = iod->sg;
    int dma_len = sg_dma_len(sg);
    u64 dma_addr = sg_dma_address(sg);
    int offset = offset_in_page(dma_addr);
    __le64 *prp_list;
    __le64 **list = iod_list(iod);
    dma_addr_t prp_dma;
    int i;

    length -= (PAGE_SIZE - offset);
    if (length <= 0)
        return total_len;

    dma_len -= (PAGE_SIZE - offset);
    if (dma_len) {
        dma_addr += (PAGE_SIZE - offset);
    } else {
        sg = sg_next(sg);
        dma_addr = sg_dma_address(sg);
        dma_len = sg_dma_len(sg);
    }

    if (length <= PAGE_SIZE) {
        iod->first_dma = dma_addr;
        return total_len;
    }

    prp_list = nubd_alloc_prp_list(dev, &prp_dma);
    if (!prp_list) {
        iod->first_dma = dma_addr;
        iod->npages = -1;
        return (total_len - length) + PAGE_SIZE;
    }
    list[0] = prp_list;
    iod->first_dma = prp_dma;
    iod->npages = 1;
    i = 0;
    for (;;) {
        if (i == PAGE_SIZE / 8) {
            __le64 *old_prp_list = prp_list;
            prp_list = nubd_alloc_prp_list(dev, &prp_dma);
            if (!prp_list)
                return total_len - length;
            list[iod->npages++] = prp_list;
            prp_list[0] = old_prp_list[i - 1];
            old_prp_list[i - 1] = cpu_to_le64(prp_dma);
            i = 1;
        }
        prp_list[i++] = cpu_to_le64(dma_addr);
        dma_len -= PAGE_SIZE;
        dma_addr += PAGE_SIZE;
        length -= PAGE_SIZE;
        if (length <= 0)
            break;
        if (dma_len > 0)
            continue;
        BUG_ON(dma_len < 0);
        sg = sg_next(sg);
        dma_addr = sg_dma_address(sg);
        dma_len = sg_dma_len(sg);
    }

    return total_len;
}

static int nvme_split_and_submit(struct bio *bio, struct nvme_queue *nvmeq,
                 int len)
{
    struct bio *split = bio_split(bio, len >> 9, GFP_ATOMIC, NULL);
    if (!split)
        return -ENOMEM;

    trace_block_split(bdev_get_queue(bio->bi_bdev), bio,
                    split->bi_iter.bi_sector);
    bio_chain(split, bio);

    if (!waitqueue_active(&nvmeq->sq_full))
        add_wait_queue(&nvmeq->sq_full, &nvmeq->sq_cong_wait);
    bio_list_add(&nvmeq->sq_cong, split);
    bio_list_add(&nvmeq->sq_cong, bio);
    wake_up(&nvmeq->sq_full);

    return 0;
}

/* NVMe scatterlists require no holes in the virtual address */
#define BIOVEC_NOT_VIRT_MERGEABLE(vec1, vec2)   ((vec2)->bv_offset || \
            (((vec1)->bv_offset + (vec1)->bv_len) % PAGE_SIZE))

static int nvme_map_bio(struct nvme_queue *nvmeq, struct nvme_iod *iod,
        struct bio *bio, enum dma_data_direction dma_dir, int psegs)
{
    struct bio_vec bvec, bvprv;
    struct bvec_iter iter;
    struct scatterlist *sg = NULL;
    int length = 0, nsegs = 0, split_len = bio->bi_iter.bi_size;
    int first = 1;

    if (nvmeq->dev->stripe_size)
        split_len = nvmeq->dev->stripe_size -
            ((bio->bi_iter.bi_sector << 9) &
             (nvmeq->dev->stripe_size - 1));

    sg_init_table(iod->sg, psegs);
    bio_for_each_segment(bvec, bio, iter) {
        if (!first && BIOVEC_PHYS_MERGEABLE(&bvprv, &bvec)) {
            sg->length += bvec.bv_len;
            sg->dma_length = sg->length;
        } else {
            if (!first && BIOVEC_NOT_VIRT_MERGEABLE(&bvprv, &bvec))
                return nvme_split_and_submit(bio, nvmeq,
                                 length);

            sg = sg ? sg + 1 : iod->sg;
            sg_set_page(sg, bvec.bv_page,
                    bvec.bv_len, bvec.bv_offset);
            sg->dma_address = page_to_phys(bvec.bv_page);
            sg->dma_address += bvec.bv_offset;
            sg->dma_length = sg->length;
            nsegs++;
        }

        if (split_len - length < bvec.bv_len)
            return nvme_split_and_submit(bio, nvmeq, split_len);
        length += bvec.bv_len;
        bvprv = bvec;
        first = 0;
    }
    iod->nents = nsegs;
    sg_mark_end(sg);
    /*
     * TODO US167552: If we need to allow EOS DMA HW access, we probably need
     * to map this in here.  Remove for time being though, doing CPU copy.
     */
#if 0
    if (dma_map_sg(nvmeq->q_dmadev, iod->sg, iod->nents, dma_dir) == 0)
        return -ENOMEM;
#endif

    BUG_ON(length != bio->bi_iter.bi_size);
    return length;
}

static int nvme_submit_discard(struct nvme_queue *nvmeq, struct nvme_ns *ns,
        struct bio *bio, struct nvme_iod *iod, int cmdid)
{
    struct nvme_dsm_range *range =
                (struct nvme_dsm_range *)iod_list(iod)[0];
    struct nvme_command *cmnd = &nvmeq->sq_cmds[nvmeq->sq_tail];

    range->cattr = cpu_to_le32(0);
    range->nlb = cpu_to_le32(bio->bi_iter.bi_size >> ns->lba_shift);
    range->slba = cpu_to_le64(nvme_block_nr(ns, bio->bi_iter.bi_sector));

    memset(cmnd, 0, sizeof(*cmnd));
    cmnd->dsm.opcode = nvme_cmd_dsm;
    cmnd->dsm.command_id = cmdid;
    cmnd->dsm.nsid = cpu_to_le32(ns->ns_id);
    cmnd->dsm.prp1 = cpu_to_le64(iod->first_dma);
    cmnd->dsm.nr = 0;
    cmnd->dsm.attributes = cpu_to_le32(NVME_DSMGMT_AD);

    if (++nvmeq->sq_tail == nvmeq->q_depth)
        nvmeq->sq_tail = 0;
    SQ_DB_NOTIFY(nvmeq->sq_tail, nvmeq->q_db);

    return 0;
}

static int nvme_submit_flush(struct nvme_queue *nvmeq, struct nvme_ns *ns,
                                int cmdid)
{
    struct nvme_command *cmnd = &nvmeq->sq_cmds[nvmeq->sq_tail];

    memset(cmnd, 0, sizeof(*cmnd));
    cmnd->common.opcode = nvme_cmd_flush;
    cmnd->common.command_id = cmdid;
    cmnd->common.nsid = cpu_to_le32(ns->ns_id);

    if (++nvmeq->sq_tail == nvmeq->q_depth)
        nvmeq->sq_tail = 0;
    SQ_DB_NOTIFY(nvmeq->sq_tail, nvmeq->q_db);

    return 0;
}

static int nvme_submit_iod(struct nvme_queue *nvmeq, struct nvme_iod *iod)
{
    struct bio *bio = iod->private;
    struct nvme_ns *ns = bio->bi_bdev->bd_disk->private_data;
    struct nvme_command *cmnd;
    int cmdid;
    u16 control;
    u32 dsmgmt;

    cmdid = alloc_cmdid(nvmeq, iod, bio_completion, NVME_IO_TIMEOUT);
    if (unlikely(cmdid < 0))
        return cmdid;

    if (bio->bi_rw & REQ_DISCARD)
        return nvme_submit_discard(nvmeq, ns, bio, iod, cmdid);
    if (bio->bi_rw & REQ_FLUSH)
        return nvme_submit_flush(nvmeq, ns, cmdid);

    control = 0;
    if (bio->bi_rw & REQ_FUA)
        control |= NVME_RW_FUA;
    if (bio->bi_rw & (REQ_FAILFAST_DEV | REQ_RAHEAD))
        control |= NVME_RW_LR;

    dsmgmt = 0;
    if (bio->bi_rw & REQ_RAHEAD)
        dsmgmt |= NVME_RW_DSM_FREQ_PREFETCH;

    cmnd = &nvmeq->sq_cmds[nvmeq->sq_tail];
    memset(cmnd, 0, sizeof(*cmnd));

    cmnd->rw.opcode = bio_data_dir(bio) ? nvme_cmd_write : nvme_cmd_read;
    cmnd->rw.command_id = cmdid;
    cmnd->rw.nsid = cpu_to_le32(ns->ns_id);
    cmnd->rw.prp1 = cpu_to_le64(sg_dma_address(iod->sg));
    cmnd->rw.prp2 = cpu_to_le64(iod->first_dma);
    cmnd->rw.slba = cpu_to_le64(nvme_block_nr(ns, bio->bi_iter.bi_sector));
    cmnd->rw.length =
        cpu_to_le16((bio->bi_iter.bi_size >> ns->lba_shift) - 1);
    cmnd->rw.control = cpu_to_le16(control);
    cmnd->rw.dsmgmt = cpu_to_le32(dsmgmt);

    if (++nvmeq->sq_tail == nvmeq->q_depth)
        nvmeq->sq_tail = 0;
    SQ_DB_NOTIFY(nvmeq->sq_tail, nvmeq->q_db);

    return 0;
}

static int nvme_split_flush_data(struct nvme_queue *nvmeq, struct bio *bio)
{
    struct bio *split = bio_clone(bio, GFP_ATOMIC);
    if (!split)
        return -ENOMEM;

    split->bi_iter.bi_size = 0;
    split->bi_phys_segments = 0;
    bio->bi_rw &= ~REQ_FLUSH;
    bio_chain(split, bio);

    if (!waitqueue_active(&nvmeq->sq_full))
        add_wait_queue(&nvmeq->sq_full, &nvmeq->sq_cong_wait);
    bio_list_add(&nvmeq->sq_cong, split);
    bio_list_add(&nvmeq->sq_cong, bio);
    wake_up_process(nvme_thread);

    return 0;
}

/*
 * Called with local interrupts disabled and the q_lock held.  May not sleep.
 */
static int nvme_submit_bio_queue(struct nvme_queue *nvmeq, struct nvme_ns *ns,
                                struct bio *bio)
{
    struct nvme_iod *iod;
    int psegs = bio_phys_segments(ns->queue, bio);
    int result;

    dev_dbg(disk_to_dev(ns->disk), "Start.");
    if ((bio->bi_rw & REQ_FLUSH) && psegs)
        return nvme_split_flush_data(nvmeq, bio);

    iod = nvme_alloc_iod(psegs, bio->bi_iter.bi_size, GFP_ATOMIC);
    if (!iod)
        return -ENOMEM;

    iod->private = bio;
    if (bio->bi_rw & REQ_DISCARD) {
        void *range;
        range = nubd_alloc_prp_list(nvmeq->dev, &iod->first_dma);
        if (!range) {
            result = -ENOMEM;
            goto free_iod;
        }
        iod_list(iod)[0] = (__le64 *)range;
        iod->npages = 0;
    } else if (psegs) {
        result = nvme_map_bio(nvmeq, iod, bio,
            bio_data_dir(bio) ? DMA_TO_DEVICE : DMA_FROM_DEVICE,
            psegs);
        if (result <= 0)
        {
            dev_warn( NUB_KDEV(nvmeq->dev), "%s: nvme_map_bio failed (result %d).", __FUNCTION__, result);
            goto free_iod;
        }
        if (nvme_setup_prps(nvmeq->dev, iod, result, GFP_ATOMIC) !=
                                result) {
            result = -ENOMEM;
            goto free_iod;
        }
        nvme_start_io_acct(bio);
    }
    result = nvme_submit_iod(nvmeq, iod);
    if (unlikely(result)) {
        dev_warn( NUB_KDEV(nvmeq->dev), "%s: submit_iod failed (result %d).", __FUNCTION__, result);
        if (!waitqueue_active(&nvmeq->sq_full))
            add_wait_queue(&nvmeq->sq_full, &nvmeq->sq_cong_wait);
        list_add_tail(&iod->node, &nvmeq->iod_bio);
    }
    dev_dbg(disk_to_dev(ns->disk), "End.");
    return 0;

 free_iod:
    nvme_free_iod(nvmeq->dev, iod);
    dev_warn( NUB_KDEV(nvmeq->dev), "%s: Ended poorly (result %d).", __FUNCTION__, result);
    return result;
}

static int nvme_process_cq(struct nvme_queue *nvmeq)
{
    u16 head, phase;

    head = nvmeq->cq_head;
    phase = nvmeq->cq_phase;

    for (;;) {
        void *ctx;
        nvme_completion_fn fn;
        struct nvme_completion cqe = nvmeq->cqes[head];
        if ((le16_to_cpu(cqe.status) & 1) != phase)
            break;
        nvmeq->sq_head = le16_to_cpu(cqe.sq_head);
        if (++head == nvmeq->q_depth) {
            head = 0;
            phase = !phase;
        }

        ctx = free_cmdid(nvmeq, cqe.command_id, &fn);
        fn(nvmeq, ctx, &cqe);
    }

    /* If the controller ignores the cq head doorbell and continuously
     * writes to the queue, it is theoretically possible to wrap around
     * the queue twice and mistakenly return IRQ_NONE.  Linux only
     * requires that 0.1% of your interrupts are handled, so this isn't
     * a big problem.
     */
    if (head == nvmeq->cq_head && phase == nvmeq->cq_phase)
        return 0;

    /*
     * The application should be polling, so technically writing the doorbell
     * value should not be needed.  We do it here to stay congruent to the
     * hardware implementation, and as it may provide some debug information
     * if things go wrong.
     */
    writel(head, nvmeq->q_db + nvmeq->dev->db_stride);

    nvmeq->cq_head = head;
    nvmeq->cq_phase = phase;

    nvmeq->cqe_seen = 1;
    return 1;
}

static void nvme_make_request(struct request_queue *q, struct bio *bio)
{
    struct nvme_ns *ns = q->queuedata;
    struct nvme_queue *nvmeq = get_nvmeq(ns->dev);
    int result = -EBUSY;

    dev_dbg(disk_to_dev(ns->disk), "Start.");
    if (!nvmeq) {
        bio_endio(bio, -EIO);
        return;
    }

    spin_lock_irq(&nvmeq->q_lock);
    if (!nvmeq->q_suspended && bio_list_empty(&nvmeq->sq_cong))
        result = nvme_submit_bio_queue(nvmeq, ns, bio);
    if (unlikely(result)) {
        dev_dbg(NUB_KDEV(nvmeq->dev),
                "submit_bio_queue failed (result %d).",
                result);
        if (!waitqueue_active(&nvmeq->sq_full))
            add_wait_queue(&nvmeq->sq_full, &nvmeq->sq_cong_wait);
        bio_list_add(&nvmeq->sq_cong, bio);
    }

    nvme_process_cq(nvmeq);
    spin_unlock_irq(&nvmeq->q_lock);
    put_nvmeq(nvmeq);
    dev_dbg(disk_to_dev(ns->disk), "End.");
}

static irqreturn_t nvme_irq(int irq, void *data)
{
    irqreturn_t result;
    struct nvme_queue *nvmeq = data;
    spin_lock(&nvmeq->q_lock);
    nvme_process_cq(nvmeq);
    result = nvmeq->cqe_seen ? IRQ_HANDLED : IRQ_NONE;
    nvmeq->cqe_seen = 0;
    spin_unlock(&nvmeq->q_lock);
    return result;
}

static void nvme_abort_command(struct nvme_queue *nvmeq, int cmdid)
{
    spin_lock_irq(&nvmeq->q_lock);
    cancel_cmdid(nvmeq, cmdid, NULL);
    spin_unlock_irq(&nvmeq->q_lock);
}

struct sync_cmd_info {
    struct task_struct *task;
    u32 result;
    int status;
};

static void sync_completion(struct nvme_queue *nvmeq, void *ctx,
                        struct nvme_completion *cqe)
{
    struct sync_cmd_info *cmdinfo = ctx;
    cmdinfo->result = le32_to_cpup(&cqe->result);
    cmdinfo->status = le16_to_cpup(&cqe->status) >> 1;
    wake_up_process(cmdinfo->task);
}

/*
 * Returns 0 on success.  If the result is negative, it's a Linux error code;
 * if the result is positive, it's an NVM Express status code
 */
static int nvme_submit_sync_cmd(struct nub_dev *dev, int q_idx,
                        struct nvme_command *cmd,
                        u32 *result, unsigned timeout)
{
    int cmdid, ret;
    struct sync_cmd_info cmdinfo;
    struct nvme_queue *nvmeq;

    nvmeq = lock_nvmeq(dev, q_idx);
    if (!nvmeq)
        return -ENODEV;

    cmdinfo.task = current;
    cmdinfo.status = -EINTR;

    cmdid = alloc_cmdid(nvmeq, &cmdinfo, sync_completion, timeout);
    if (cmdid < 0) {
        unlock_nvmeq(nvmeq);
        return cmdid;
    }
    cmd->common.command_id = cmdid;

    set_current_state(TASK_KILLABLE);
    ret = nvme_submit_cmd(nvmeq, cmd);
    if (ret) {
        free_cmdid(nvmeq, cmdid, NULL);
        unlock_nvmeq(nvmeq);
        set_current_state(TASK_RUNNING);
        return ret;
    }
    unlock_nvmeq(nvmeq);
    schedule_timeout(timeout);

    if (cmdinfo.status == -EINTR) {
        nvmeq = lock_nvmeq(dev, q_idx);
        if (nvmeq) {
            nvme_abort_command(nvmeq, cmdid);
            unlock_nvmeq(nvmeq);
        }
        return -EINTR;
    }

    if (result)
        *result = cmdinfo.result;

    return cmdinfo.status;
}

static int nvme_submit_async_cmd(struct nvme_queue *nvmeq,
            struct nvme_command *cmd,
            struct async_cmd_info *cmdinfo, unsigned timeout)
{
    int cmdid;

    cmdid = alloc_cmdid_killable(nvmeq, cmdinfo, async_completion, timeout);
    if (cmdid < 0)
        return cmdid;
    cmdinfo->status = -EINTR;
    cmd->common.command_id = cmdid;
    return nvme_submit_cmd(nvmeq, cmd);
}

int nvme_submit_admin_cmd(struct nub_dev *dev, struct nvme_command *cmd,
                                u32 *result)
{
    return nvme_submit_sync_cmd(dev, 0, cmd, result, ADMIN_TIMEOUT);
}

int nvme_submit_io_cmd(struct nub_dev *dev, struct nvme_command *cmd,
                                u32 *result)
{
    return nvme_submit_sync_cmd(dev, smp_processor_id() + 1, cmd, result,
                            NVME_IO_TIMEOUT);
}

static int nvme_submit_admin_cmd_async(struct nub_dev *dev,
        struct nvme_command *cmd, struct async_cmd_info *cmdinfo)
{
    return nvme_submit_async_cmd(raw_nvmeq(dev, 0), cmd, cmdinfo,
                                ADMIN_TIMEOUT);
}

static int adapter_delete_queue(struct nub_dev *dev, u8 opcode, u16 id)
{
    int status;
    struct nvme_command c;

    memset(&c, 0, sizeof(c));
    c.delete_queue.opcode = opcode;
    c.delete_queue.qid = cpu_to_le16(id);

    status = nvme_submit_admin_cmd(dev, &c, NULL);
    if (status)
        return -EIO;
    return 0;
}

static int adapter_alloc_cq(struct nub_dev *dev, u16 qid,
                        struct nvme_queue *nvmeq)
{
    int status;
    struct nvme_command c;
    int flags = NVME_QUEUE_PHYS_CONTIG | NVME_CQ_IRQ_ENABLED;

    memset(&c, 0, sizeof(c));
    c.create_cq.opcode = nvme_admin_create_cq;
    c.create_cq.prp1 = cpu_to_le64(nvmeq->cq_dma_addr);
    c.create_cq.cqid = cpu_to_le16(qid);
    c.create_cq.qsize = cpu_to_le16(nvmeq->q_depth - 1);
    c.create_cq.cq_flags = cpu_to_le16(flags);
    /* irq_vector is not relevant to user space driver. */
    c.create_cq.irq_vector = cpu_to_le16(0);

    status = nvme_submit_admin_cmd(dev, &c, NULL);
    if (status)
        return -EIO;
    return 0;
}

static int adapter_alloc_sq(struct nub_dev *dev, u16 qid,
                        struct nvme_queue *nvmeq)
{
    int status;
    struct nvme_command c;
    int flags = NVME_QUEUE_PHYS_CONTIG | NVME_SQ_PRIO_MEDIUM;

    memset(&c, 0, sizeof(c));
    c.create_sq.opcode = nvme_admin_create_sq;
    c.create_sq.prp1 = cpu_to_le64(nvmeq->sq_dma_addr);
    c.create_sq.sqid = cpu_to_le16(qid);
    c.create_sq.qsize = cpu_to_le16(nvmeq->q_depth - 1);
    c.create_sq.sq_flags = cpu_to_le16(flags);
    c.create_sq.cqid = cpu_to_le16(qid);

    status = nvme_submit_admin_cmd(dev, &c, NULL);
    if (status)
        return -EIO;
    return 0;
}

static int adapter_delete_cq(struct nub_dev *dev, u16 cqid)
{
    return adapter_delete_queue(dev, nvme_admin_delete_cq, cqid);
}

static int adapter_delete_sq(struct nub_dev *dev, u16 sqid)
{
    return adapter_delete_queue(dev, nvme_admin_delete_sq, sqid);
}

int nvme_identify(struct nub_dev *dev, unsigned nsid, unsigned cns,
                            dma_addr_t dma_addr)
{
    struct nvme_command c;

    memset(&c, 0, sizeof(c));
    c.identify.opcode = nvme_admin_identify;
    c.identify.nsid = cpu_to_le32(nsid);
    c.identify.prp1 = cpu_to_le64(dma_addr);
    c.identify.cns = cpu_to_le32(cns);

    return nvme_submit_admin_cmd(dev, &c, NULL);
}

int nvme_get_features(struct nub_dev *dev, unsigned fid, unsigned nsid,
                    dma_addr_t dma_addr, u32 *result)
{
    struct nvme_command c;

    memset(&c, 0, sizeof(c));
    c.features.opcode = nvme_admin_get_features;
    c.features.nsid = cpu_to_le32(nsid);
    c.features.prp1 = cpu_to_le64(dma_addr);
    c.features.fid = cpu_to_le32(fid);

    return nvme_submit_admin_cmd(dev, &c, result);
}

int nvme_set_features(struct nub_dev *dev, unsigned fid, unsigned dword11,
                    dma_addr_t dma_addr, u32 *result)
{
    struct nvme_command c;

    memset(&c, 0, sizeof(c));
    c.features.opcode = nvme_admin_set_features;
    c.features.prp1 = cpu_to_le64(dma_addr);
    c.features.fid = cpu_to_le32(fid);
    c.features.dword11 = cpu_to_le32(dword11);

    return nvme_submit_admin_cmd(dev, &c, result);
}

/**
 * nvme_abort_cmd - Attempt aborting a command
 * @cmdid: Command id of a timed out IO
 * @queue: The queue with timed out IO
 *
 * Schedule controller reset if the command was already aborted once before and
 * still hasn't been returned to the driver, or if this is the admin queue.
 */
static void nvme_abort_cmd(int cmdid, struct nvme_queue *nvmeq)
{
    int a_cmdid;
    struct nvme_command cmd;
    struct nub_dev *dev = nvmeq->dev;
    struct device *ctrl_device = NUB_KDEV(dev);
    struct nvme_cmd_info *info = nvme_cmd_info(nvmeq);
    struct nvme_queue *adminq;

    if (!nvmeq->qid || info[cmdid].aborted) {
        if (work_busy(&dev->reset_work))
            return;
        list_del_init(&dev->node);
        dev_warn(ctrl_device,
            "I/O %d QID %d timeout, reset controller\n", cmdid,
                                nvmeq->qid);
        dev->reset_workfn = nvme_reset_failed_dev;
        queue_work(nvme_workq, &dev->reset_work);
        return;
    }

    if (!dev->abort_limit)
        return;

    adminq = rcu_dereference(dev->queues[0]);
    a_cmdid = alloc_cmdid(adminq, CMD_CTX_ABORT, special_completion,
                                ADMIN_TIMEOUT);
    if (a_cmdid < 0)
        return;

    memset(&cmd, 0, sizeof(cmd));
    cmd.abort.opcode = nvme_admin_abort_cmd;
    cmd.abort.cid = cmdid;
    cmd.abort.sqid = cpu_to_le16(nvmeq->qid);
    cmd.abort.command_id = a_cmdid;

    --dev->abort_limit;
    info[cmdid].aborted = 1;
    info[cmdid].timeout = jiffies + ADMIN_TIMEOUT;

    dev_warn(ctrl_device, "Aborting I/O %d QID %d\n", cmdid,
                            nvmeq->qid);
    nvme_submit_cmd(adminq, &cmd);
}

/**
 * nvme_cancel_ios - Cancel outstanding I/Os
 * @queue: The queue to cancel I/Os on
 * @timeout: True to only cancel I/Os which have timed out
 */
static void nvme_cancel_ios(struct nvme_queue *nvmeq, bool timeout)
{
    int depth = nvmeq->q_depth - 1;
    struct nvme_cmd_info *info = nvme_cmd_info(nvmeq);
    unsigned long now = jiffies;
    int cmdid;

    for_each_set_bit(cmdid, nvmeq->cmdid_data, depth) {
        void *ctx;
        nvme_completion_fn fn;
        static struct nvme_completion cqe = {
            .status = cpu_to_le16(NVME_SC_ABORT_REQ << 1),
        };

        if (timeout && !time_after(now, info[cmdid].timeout))
            continue;
        if (info[cmdid].ctx == CMD_CTX_CANCELLED)
            continue;
        if (timeout && nvmeq->dev->initialized) {
            nvme_abort_cmd(cmdid, nvmeq);
            continue;
        }
        dev_warn(nvmeq->q_dmadev, "Cancelling I/O %d QID %d\n", cmdid,
                                nvmeq->qid);
        ctx = cancel_cmdid(nvmeq, cmdid, &fn);
        fn(nvmeq, ctx, &cqe);
    }
}

static void nvme_free_queue(struct rcu_head *r)
{
    struct nvme_queue *nvmeq = container_of(r, struct nvme_queue, r_head);

    spin_lock_irq(&nvmeq->q_lock);
    while (bio_list_peek(&nvmeq->sq_cong)) {
        struct bio *bio = bio_list_pop(&nvmeq->sq_cong);
        bio_endio(bio, -EIO);
    }
    while (!list_empty(&nvmeq->iod_bio)) {
        static struct nvme_completion cqe = {
            .status = cpu_to_le16(
                (NVME_SC_ABORT_REQ | NVME_SC_DNR) << 1),
        };
        struct nvme_iod *iod = list_first_entry(&nvmeq->iod_bio,
                            struct nvme_iod,
                            node);
        list_del(&iod->node);
        bio_completion(nvmeq, iod, &cqe);
    }
    spin_unlock_irq(&nvmeq->q_lock);

    if (nvmeq->qid)
        free_cpumask_var(nvmeq->cpu_mask);
    kfree(nvmeq);
}

static void nvme_free_queues(struct nub_dev *dev, int lowest)
{
    int i;

    for (i = dev->queue_count - 1; i >= lowest; i--) {
        struct nvme_queue *nvmeq = raw_nvmeq(dev, i);
        rcu_assign_pointer(dev->queues[i], NULL);
        call_rcu(&nvmeq->r_head, nvme_free_queue);
        dev->queue_count--;
    }
}

/**
 * nvme_suspend_queue - put queue into suspended state
 * @nvmeq - queue to suspend
 *
 * Returns 1 if already suspended, 0 otherwise.
 */
static int nvme_suspend_queue(struct nvme_queue *nvmeq)
{
    spin_lock_irq(&nvmeq->q_lock);
    if (nvmeq->q_suspended) {
        spin_unlock_irq(&nvmeq->q_lock);
        return 1;
    }
    nvmeq->q_suspended = 1;
    nvmeq->dev->online_queues--;
    spin_unlock_irq(&nvmeq->q_lock);

    return 0;
}

static void nvme_clear_queue(struct nvme_queue *nvmeq)
{
    spin_lock_irq(&nvmeq->q_lock);
    nvme_process_cq(nvmeq);
    nvme_cancel_ios(nvmeq, false);
    spin_unlock_irq(&nvmeq->q_lock);
}

static void nvme_disable_queue(struct nub_dev *dev, int qid)
{
    struct nvme_queue *nvmeq = raw_nvmeq(dev, qid);

    if (!nvmeq)
        return;
    if (nvme_suspend_queue(nvmeq))
        return;

    /* Don't tell the adapter to delete the admin queue. */
    if (qid) {
        adapter_delete_sq(dev, qid);
        adapter_delete_cq(dev, qid);
    }
    nvme_clear_queue(nvmeq);
}


/*
 * Fixed-size block allocator for PRP list resources (per-queue).
 * I couldn't find anything that would operate well with the CMB,
 * so I wrote a simple one here.
 */
static void* nubd_alloc_prp_list(struct nub_dev* dev, dma_addr_t *handle)
{
    unsigned long idx = find_first_zero_bit(dev->free_prpl_bitmap, dev->prp_pool_size);
    if (idx >=  dev->prp_pool_size)
        return 0;

    while (test_and_set_bit(idx, dev->free_prpl_bitmap))
    {
        idx = find_first_zero_bit(dev->free_prpl_bitmap, dev->prp_pool_size);
        if (idx >=  dev->prp_pool_size)
            return 0;
    }

    *handle = dev->prpl_dma_addr + (idx * sizeof(struct prp_list));
    return &(dev->prpl[idx]);
}

static void nubd_free_prp_list(struct nub_dev* dev, void* vaddr)
{
    unsigned long idx = ((uintptr_t)vaddr - (uintptr_t)dev->prpl) / sizeof(struct prp_list);
    clear_bit(idx, dev->free_prpl_bitmap);
}


static void cq_update_workfn(struct work_struct *work)
{
    struct nvme_queue *cq = container_of(work, struct nvme_queue, cq_update_work);
    nvme_irq(0, cq);
}


/*
 * Allocates a CQ/SQ pair given the requested depths and qid.
 * This currently assumes that max IO size is 2MB-PAGE_SIZE or less so we can
 * fit all PRP entries in a single PRP list on x86 systems.
 */
static struct nvme_queue *nvme_alloc_queue(struct nub_dev *dev, int qid,
                            int depth, int vector)
{
    unsigned extra = nvme_queue_extra(depth);
    struct nvme_queue *nvmeq;
    struct workqueue_struct *wq;
    void*   req_end =  (char*)dev->cmb_pbase + dev->cmb_unused_offset
                        + CQ_SIZE(depth) + SQ_SIZE(depth);

    dev_dbg(NUB_KDEV(dev), "Start. (qid %d)\n", qid);
    dev_dbg(NUB_KDEV(dev), "CQ size:0x%lx, SQ size:0x%lx, req_end:0x%p\n",
               CQ_SIZE(depth), SQ_SIZE(depth), req_end);
    /*
     * Sanity check io queues to make sure we have enough room,
     * which we should have if the previous setup was done right.
     * We don't have a prpl address yet when we init the admin queue,
     * so we do not check it then.
     */
    if (qid && ((dma_addr_t)req_end >= dev->prpl_dma_addr))
        return NULL;

    nvmeq = kzalloc(sizeof(*nvmeq) + extra, GFP_KERNEL);
    if (!nvmeq)
        return NULL;

    /* Figure DMA addresses for the queues. */
    nvmeq->sq_dma_addr = (uintptr_t)dev->cmb_pbase + dev->cmb_unused_offset;
    nvmeq->cq_dma_addr = nvmeq->sq_dma_addr + SQ_SIZE(depth);

    nvmeq->cqes = KV_CMB_ADDR(dev, nvmeq->cq_dma_addr);
    memset((void *)nvmeq->cqes, 0, CQ_SIZE(depth));

    nvmeq->sq_cmds = KV_CMB_ADDR(dev, nvmeq->sq_dma_addr);

    /* Update offset to next available SQ entry-size aligned page address. */
    dev->cmb_unused_offset += CQ_SIZE(depth) + SQ_SIZE(depth);
    dev->cmb_unused_offset = round_up(dev->cmb_unused_offset, sizeof(struct nvme_command));

    dev_dbg(NUB_KDEV(dev), "QID %d: SQ at 0x%p\n", qid, (void*)nvmeq->sq_dma_addr);
    dev_dbg(NUB_KDEV(dev), "QID %d: CQ at 0x%p\n", qid, (void*)nvmeq->cq_dma_addr);
    dev_dbg(NUB_KDEV(dev), "QID %d: unused area offset 0x%zx\n", qid, dev->cmb_unused_offset);

    if (qid && !zalloc_cpumask_var(&nvmeq->cpu_mask, GFP_KERNEL))
        goto free_nvmeq;

    nvmeq->q_dmadev = NUB_KDEV(dev);
    nvmeq->dev = dev;
    snprintf(nvmeq->irqname, sizeof(nvmeq->irqname), "nvme%dq%d",
            dev->instance, qid);

    /*
     * The intention is that the user space part of the driver will signal us via the
     * workqueue whenever there is work to do.  We might also consider some poll
     * cycle every few seconds to check the queue for pending work in case the
     * message somehow gets missed.
     */
    wq = create_singlethread_workqueue(nvmeq->irqname);
    if (!wq)
        goto free_cpumask;

    spin_lock_init(&nvmeq->q_lock);
    nvmeq->cq_head = 0;
    nvmeq->cq_phase = 1;
    init_waitqueue_head(&nvmeq->sq_full);
    init_waitqueue_entry(&nvmeq->sq_cong_wait, nvme_thread);
    bio_list_init(&nvmeq->sq_cong);
    INIT_LIST_HEAD(&nvmeq->iod_bio);
    nvmeq->q_db = &dev->dbs[qid * 2 * dev->db_stride];
    nvmeq->q_depth = depth;
    nvmeq->io_workqueue = wq;
    INIT_WORK( &nvmeq->cq_update_work, cq_update_workfn );

    nvmeq->qid = qid;
    nvmeq->q_suspended = 1;
    dev->queue_count++;
    rcu_assign_pointer(dev->queues[qid], nvmeq);

    dev_dbg(NUB_KDEV(dev), "End.\n");
    return nvmeq;

 free_cpumask:
    free_cpumask_var(nvmeq->cpu_mask);
 free_nvmeq:
    kfree(nvmeq);
    return NULL;
}


static void nvme_init_queue(struct nvme_queue *nvmeq, u16 qid)
{
    struct nub_dev *dev = nvmeq->dev;
    unsigned extra = nvme_queue_extra(nvmeq->q_depth);

    nvmeq->sq_tail = 0;
    nvmeq->cq_head = 0;
    nvmeq->cq_phase = 1;
    nvmeq->q_db = &dev->dbs[qid * 2 * dev->db_stride];
    memset(nvmeq->cmdid_data, 0, extra);
    memset((void *)nvmeq->cqes, 0, CQ_SIZE(nvmeq->q_depth));
    nvme_cancel_ios(nvmeq, false);
    nvmeq->q_suspended = 0;
    dev->online_queues++;
}

static int nvme_create_queue(struct nvme_queue *nvmeq, int qid)
{
    struct nub_dev *dev = nvmeq->dev;
    int result;

    result = adapter_alloc_cq(dev, qid, nvmeq);
    if (result < 0)
        return result;

    result = adapter_alloc_sq(dev, qid, nvmeq);
    if (result < 0)
        goto release_cq;


    spin_lock_irq(&nvmeq->q_lock);
    nvme_init_queue(nvmeq, qid);
    spin_unlock_irq(&nvmeq->q_lock);

    return result;

 release_cq:
    adapter_delete_cq(dev, qid);
    return result;
}

static int nvme_wait_ready(struct nub_dev *dev, u64 cap, bool enabled)
{
    unsigned long timeout;
    u32 bit = enabled ? NVME_CSTS_RDY : 0;

    timeout = ((NVME_CAP_TIMEOUT(cap) + 1) * HZ / 2) + jiffies;

    while ((readl(&dev->bar->csts) & NVME_CSTS_RDY) != bit) {
        msleep(100);
        if (fatal_signal_pending(current))
            return -EINTR;
        if (time_after(jiffies, timeout)) {
            dev_err(NUB_KDEV(dev),
                "Device not ready; aborting %s\n", enabled ?
                        "initialisation" : "reset");
            return -ENODEV;
        }
    }

    return 0;
}

/*
 * If the device has been passed off to us in an enabled state, just clear
 * the enabled bit.  The spec says we should set the 'shutdown notification
 * bits', but doing so may cause the device to complete commands to the
 * admin queue ... and we don't know what memory that might be pointing at!
 */
static int nvme_disable_ctrl(struct nub_dev *dev, u64 cap)
{
    u32 cc = readl(&dev->bar->cc);

    if (cc & NVME_CC_ENABLE)
        writel(cc & ~NVME_CC_ENABLE, &dev->bar->cc);
    return nvme_wait_ready(dev, cap, false);
}

static int nvme_enable_ctrl(struct nub_dev *dev, u64 cap)
{
    return nvme_wait_ready(dev, cap, true);
}

static int nvme_configure_admin_queue(struct nub_dev *dev)
{
    int result;
    u32 aqa;
    u64 cap = readq(&dev->bar->cap);
    struct nvme_queue *nvmeq;

    dev_dbg(NUB_KDEV(dev), "Start...\n");
    result = nvme_disable_ctrl(dev, cap);
    if (result < 0)
        return result;

    nvmeq = raw_nvmeq(dev, 0);
    if (!nvmeq) {
        nvmeq = nvme_alloc_queue(dev, 0, DEFAULT_ADMIN_QDEPTH, 0);
        if (!nvmeq)
            return -ENOMEM;
    }

    aqa = nvmeq->q_depth - 1;
    aqa |= aqa << 16;

    dev->ctrl_config = NVME_CC_ENABLE | NVME_CC_CSS_NVM;
    dev->ctrl_config |= (PAGE_SHIFT - 12) << NVME_CC_MPS_SHIFT;
    dev->ctrl_config |= NVME_CC_ARB_RR | NVME_CC_SHN_NONE;
    dev->ctrl_config |= NVME_CC_IOSQES | NVME_CC_IOCQES;

    writel(aqa, &dev->bar->aqa);
    writeq(nvmeq->sq_dma_addr, &dev->bar->asq);
    writeq(nvmeq->cq_dma_addr, &dev->bar->acq);
    writel(dev->ctrl_config, &dev->bar->cc);

    result = nvme_enable_ctrl(dev, cap);
    if (result)
        return result;

    dev_dbg(NUB_KDEV(dev), "AdminQ 0x%p starting\n", nvmeq);
    spin_lock_irq(&nvmeq->q_lock);
    nvme_init_queue(nvmeq, 0);
    spin_unlock_irq(&nvmeq->q_lock);
    dev_dbg(NUB_KDEV(dev), "End.\n");
    return 0;
}

struct nvme_iod *nvme_map_user_pages(struct nub_dev *dev, int write,
                unsigned long addr, unsigned length)
{
    int i, err, count, offset;
    struct scatterlist *sg;
    struct page **pages;
    struct nvme_iod *iod;

    if (addr & 3)
        return ERR_PTR(-EINVAL);
    if (!length || length > INT_MAX - PAGE_SIZE)
        return ERR_PTR(-EINVAL);

    offset = offset_in_page(addr);
    count = DIV_ROUND_UP(offset + length, PAGE_SIZE);
    pages = kcalloc(count, sizeof(*pages), GFP_KERNEL);
    if (!pages)
        return ERR_PTR(-ENOMEM);

    err = get_user_pages_fast(addr, count, 1, pages);
    if (err < count) {
        count = err;
        err = -EFAULT;
        goto put_pages;
    }

    err = -ENOMEM;
    iod = nvme_alloc_iod(count, length, GFP_KERNEL);
    if (!iod)
        goto put_pages;

    sg = iod->sg;
    sg_init_table(sg, count);
    for (i = 0; i < count; i++) {
        sg_set_page(&sg[i], pages[i],
                min_t(unsigned, length, PAGE_SIZE - offset),
                offset);
        length -= (PAGE_SIZE - offset);
        offset = 0;
    }
    sg_mark_end(&sg[i - 1]);
    iod->nents = count;

    /*
     * TODO US167552: If we need to allow EOS DMA HW access, we probably need
     * to map this in here.  Remove for time being though, doing CPU copy.
     */
#if 0
    nents = dma_map_sg(&dev->pci_dev->dev, sg, count,
                write ? DMA_TO_DEVICE : DMA_FROM_DEVICE);
    if (!nents)
        goto free_iod;
#endif

    kfree(pages);
    return iod;

 put_pages:
    for (i = 0; i < count; i++)
        put_page(pages[i]);
    kfree(pages);
    return ERR_PTR(err);
}

void nvme_unmap_user_pages(struct nub_dev *dev, int write,
            struct nvme_iod *iod)
{
    int i;

    dma_unmap_sg(&dev->pci_dev->dev, iod->sg, iod->nents,
                write ? DMA_TO_DEVICE : DMA_FROM_DEVICE);

    for (i = 0; i < iod->nents; i++)
        put_page(sg_page(&iod->sg[i]));
}
#endif

static int nvme_ioctl(struct block_device *bdev, fmode_t mode, unsigned int cmd,
                            unsigned long arg)
{
#ifndef DISABLE_NUBD_DRIVER
    struct nvme_ns *ns = bdev->bd_disk->private_data;

    dev_dbg(disk_to_dev(bdev->bd_disk), "Start.");
    switch (cmd) {
    case NUBD_IOCTL_NSID:
        force_successful_syscall_return();
        return ns->ns_id;
    case SG_GET_VERSION_NUM:
        return nvme_sg_get_version_num((void __user *)arg);
    case SG_IO:
        return nvme_sg_io(ns, (void __user *)arg);
    default:
        return -ENOTTY;
    }
#else
    return -EINVAL;
#endif
}

#ifdef CONFIG_COMPAT
static int nvme_compat_ioctl(struct block_device *bdev, fmode_t mode,
                    unsigned int cmd, unsigned long arg)
{
    struct nvme_ns *ns = bdev->bd_disk->private_data;

    dev_dbg(disk_to_dev(bdev->bd_disk), "Start.");
    switch (cmd) {
    case SG_IO:
        return nvme_sg_io32(ns, arg);
    }
    return nvme_ioctl(bdev, mode, cmd, arg);
}
#else
#define nvme_compat_ioctl   NULL
#endif

#ifndef DISABLE_NUBD_DRIVER
static int nvme_open(struct block_device *bdev, fmode_t mode)
{
    struct nvme_ns *ns = bdev->bd_disk->private_data;
    struct nub_dev *dev = ns->dev;

    dev_dbg(disk_to_dev(bdev->bd_disk), "Start.");
    kref_get(&dev->kref);
    return 0;
}

static void nvme_free_dev(struct kref *kref);

static void nvme_release(struct gendisk *disk, fmode_t mode)
{
    struct nvme_ns *ns = disk->private_data;
    struct nub_dev *dev = ns->dev;

    dev_dbg(disk_to_dev(disk), "Start.");
    kref_put(&dev->kref, nvme_free_dev);
}

static int nvme_getgeo(struct block_device *bd, struct hd_geometry *geo)
{
    dev_dbg(disk_to_dev(bd->bd_disk), "Start.");
    /* some standard values */
    geo->heads = 1 << 6;
    geo->sectors = 1 << 5;
    geo->cylinders = get_capacity(bd->bd_disk) >> 11;
    return 0;
}

static const struct block_device_operations nvme_fops = {
    .owner      = THIS_MODULE,
    .ioctl      = nvme_ioctl,
    .compat_ioctl   = nvme_compat_ioctl,
    .open       = nvme_open,
    .release    = nvme_release,
    .getgeo     = nvme_getgeo,
};

static void nvme_resubmit_iods(struct nvme_queue *nvmeq)
{
    struct nvme_iod *iod, *next;

    list_for_each_entry_safe(iod, next, &nvmeq->iod_bio, node) {
        if (unlikely(nvme_submit_iod(nvmeq, iod)))
            break;
        list_del(&iod->node);
        if (bio_list_empty(&nvmeq->sq_cong) &&
                        list_empty(&nvmeq->iod_bio))
            remove_wait_queue(&nvmeq->sq_full,
                        &nvmeq->sq_cong_wait);
    }
}

static void nvme_resubmit_bios(struct nvme_queue *nvmeq)
{
    while (bio_list_peek(&nvmeq->sq_cong)) {
        struct bio *bio = bio_list_pop(&nvmeq->sq_cong);
        struct nvme_ns *ns = bio->bi_bdev->bd_disk->private_data;

        if (bio_list_empty(&nvmeq->sq_cong) &&
                        list_empty(&nvmeq->iod_bio))
            remove_wait_queue(&nvmeq->sq_full,
                            &nvmeq->sq_cong_wait);
        if (nvme_submit_bio_queue(nvmeq, ns, bio)) {
            if (!waitqueue_active(&nvmeq->sq_full))
                add_wait_queue(&nvmeq->sq_full,
                            &nvmeq->sq_cong_wait);
            bio_list_add_head(&nvmeq->sq_cong, bio);
            break;
        }
    }
}

static int nvme_kthread(void *data)
{
    struct nub_dev *dev, *next;

    while (!kthread_should_stop()) {
        set_current_state(TASK_INTERRUPTIBLE);
        spin_lock(&dev_list_lock);
        list_for_each_entry_safe(dev, next, &dev_list, node) {
            int i;
            if (!dev->initialized)
                continue;
            if (readl(&dev->bar->csts) & NVME_CSTS_CFS) {
                if (work_busy(&dev->reset_work))
                    continue;
                list_del_init(&dev->node);
                dev_warn(NUB_KDEV(dev),
                    "Failed status, reset controller\n");
                dev->reset_workfn = nvme_reset_failed_dev;
                queue_work(nvme_workq, &dev->reset_work);
                continue;
            }
            rcu_read_lock();
            for (i = 0; i < dev->queue_count; i++) {
                struct nvme_queue *nvmeq =
                        rcu_dereference(dev->queues[i]);
                if (!nvmeq)
                    continue;
                spin_lock_irq(&nvmeq->q_lock);
                if (nvmeq->q_suspended)
                    goto unlock;
                nvme_process_cq(nvmeq);
                nvme_cancel_ios(nvmeq, true);
                nvme_resubmit_bios(nvmeq);
                nvme_resubmit_iods(nvmeq);
 unlock:
                spin_unlock_irq(&nvmeq->q_lock);
            }
            rcu_read_unlock();
        }
        spin_unlock(&dev_list_lock);
        schedule_timeout(round_jiffies_relative(HZ));
    }
    return 0;
}

static void nvme_config_discard(struct nvme_ns *ns)
{
    u32 logical_block_size = queue_logical_block_size(ns->queue);
    ns->queue->limits.discard_zeroes_data = 0;
    ns->queue->limits.discard_alignment = logical_block_size;
    ns->queue->limits.discard_granularity = logical_block_size;
    ns->queue->limits.max_discard_sectors = 0xffffffff;
    queue_flag_set_unlocked(QUEUE_FLAG_DISCARD, ns->queue);
}

static struct nvme_ns *nvme_alloc_ns(struct nub_dev *dev, unsigned nsid,
            struct nvme_id_ns *id, struct nvme_lba_range_type *rt)
{
    struct nvme_ns *ns;
    struct gendisk *disk;
    int lbaf;

    if (rt->attributes & NVME_LBART_ATTRIB_HIDE)
        return NULL;

    ns = kzalloc(sizeof(*ns), GFP_KERNEL);
    if (!ns)
        return NULL;
    ns->queue = blk_alloc_queue(GFP_KERNEL);
    if (!ns->queue)
        goto out_free_ns;
    ns->queue->queue_flags = QUEUE_FLAG_DEFAULT;
    queue_flag_set_unlocked(QUEUE_FLAG_NOMERGES, ns->queue);
    queue_flag_set_unlocked(QUEUE_FLAG_NONROT, ns->queue);
    blk_queue_make_request(ns->queue, nvme_make_request);
    ns->dev = dev;
    ns->queue->queuedata = ns;

    disk = alloc_disk(1);
    if (!disk)
        goto out_free_queue;
    ns->ns_id = nsid;
    ns->disk = disk;
    lbaf = id->flbas & 0xf;
    ns->lba_shift = id->lbaf[lbaf].ds;
    ns->ms = le16_to_cpu(id->lbaf[lbaf].ms);
    blk_queue_logical_block_size(ns->queue, 1 << ns->lba_shift);
    if (dev->max_hw_sectors)
    {
        dev_dbg(NUB_KDEV(dev), "setting max_hw_sectors:x%x\n", dev->max_hw_sectors);
        blk_queue_max_hw_sectors(ns->queue, dev->max_hw_sectors);
    }
    if (dev->vwc & NVME_CTRL_VWC_PRESENT)
        blk_queue_flush(ns->queue, REQ_FLUSH | REQ_FUA);

    disk->major = nubd_major;
    disk->first_minor = nsid;
    disk->fops = &nvme_fops;
    disk->private_data = ns;
    disk->queue = ns->queue;
    disk->driverfs_dev = NUB_KDEV(dev);
    disk->flags = GENHD_FL_EXT_DEVT;
    sprintf(disk->disk_name, "eos-lun/%d", nsid);
    set_capacity(disk, le64_to_cpup(&id->ncap) << (ns->lba_shift - 9));

    if (dev->oncs & NVME_CTRL_ONCS_DSM)
        nvme_config_discard(ns);

    return ns;

 out_free_queue:
    blk_cleanup_queue(ns->queue);
 out_free_ns:
    kfree(ns);
    return NULL;
}

static int nvme_find_closest_node(int node)
{
    int n, val, min_val = INT_MAX, best_node = node;

    for_each_online_node(n) {
        if (n == node)
            continue;
        val = node_distance(node, n);
        if (val < min_val) {
            min_val = val;
            best_node = n;
        }
    }
    return best_node;
}

static void nvme_set_queue_cpus(cpumask_t *qmask, struct nvme_queue *nvmeq,
                                int count)
{
    int cpu;
    for_each_cpu(cpu, qmask) {
        if (cpumask_weight(nvmeq->cpu_mask) >= count)
            break;
        if (!cpumask_test_and_set_cpu(cpu, nvmeq->cpu_mask))
            *per_cpu_ptr(nvmeq->dev->io_queue, cpu) = nvmeq->qid;
    }
}

static void nvme_add_cpus(cpumask_t *mask, const cpumask_t *unassigned_cpus,
    const cpumask_t *new_mask, struct nvme_queue *nvmeq, int cpus_per_queue)
{
    int next_cpu;
    for_each_cpu(next_cpu, new_mask) {
        cpumask_or(mask, mask, get_cpu_mask(next_cpu));
        cpumask_or(mask, mask, topology_thread_cpumask(next_cpu));
        cpumask_and(mask, mask, unassigned_cpus);
        nvme_set_queue_cpus(mask, nvmeq, cpus_per_queue);
    }
}

static void nvme_create_io_queues(struct nub_dev *dev)
{
    unsigned i, max;

    dev_dbg(NUB_KDEV(dev), "Start.\n");
    max = min(dev->max_qid, num_online_cpus());
    for (i = dev->queue_count; i <= max; i++)
        if (!nvme_alloc_queue(dev, i, dev->q_depth, i - 1))
            break;

    max = min(dev->queue_count - 1, num_online_cpus());
    for (i = dev->online_queues; i <= max; i++)
    {
        dev_dbg(NUB_KDEV(dev), "create queue %d...\n", i);
        if (nvme_create_queue(raw_nvmeq(dev, i), i))
            break;
    }
    dev_dbg(NUB_KDEV(dev), "End.\n");
}

/*
 * If there are fewer queues than online cpus, this will try to optimally
 * assign a queue to multiple cpus by grouping cpus that are "close" together:
 * thread siblings, core, socket, closest node, then whatever else is
 * available.
 *
 * To avoid running threads on the user space driver core(s), pin the application
 * to high-numbered cores and only advertise queues less than or equal
 * to the number of cores left over.
 */
static void nvme_assign_io_queues(struct nub_dev *dev)
{
    unsigned cpu, cpus_per_queue, queues, remainder, i;
    cpumask_var_t unassigned_cpus;

    dev_dbg(NUB_KDEV(dev), "Start...\n");
    nvme_create_io_queues(dev);

    queues = min(dev->online_queues - 1, num_online_cpus());
    if (!queues)
        return;

    cpus_per_queue = num_online_cpus() / queues;
    remainder = queues - (num_online_cpus() - queues * cpus_per_queue);

    dev_dbg(NUB_KDEV(dev), "Trying to init %d queues...\n", queues);
    if (!alloc_cpumask_var(&unassigned_cpus, GFP_KERNEL))
        return;

    cpumask_copy(unassigned_cpus, cpu_online_mask);
    cpu = cpumask_first(unassigned_cpus);
    for (i = 1; i <= queues; i++) {
        struct nvme_queue *nvmeq = lock_nvmeq(dev, i);
        cpumask_t mask;
        dev_dbg(NUB_KDEV(dev), "Initializing queue %d.\n", i);

        cpumask_clear(nvmeq->cpu_mask);
        if (!cpumask_weight(unassigned_cpus)) {
            unlock_nvmeq(nvmeq);
            break;
        }

        mask = *get_cpu_mask(cpu);
        nvme_set_queue_cpus(&mask, nvmeq, cpus_per_queue);
        if (cpus_weight(mask) < cpus_per_queue)
            nvme_add_cpus(&mask, unassigned_cpus,
                topology_thread_cpumask(cpu),
                nvmeq, cpus_per_queue);
        if (cpus_weight(mask) < cpus_per_queue)
            nvme_add_cpus(&mask, unassigned_cpus,
                topology_core_cpumask(cpu),
                nvmeq, cpus_per_queue);
        /* NOTE: In our case, we should never hit the following 3 cases
         *      Since all of our system are single-chip, multi-core
         *      processors.  If we had multiple chips though this should
         *      work fine.
         */
        if (cpus_weight(mask) < cpus_per_queue)
            nvme_add_cpus(&mask, unassigned_cpus,
                cpumask_of_node(cpu_to_node(cpu)),
                nvmeq, cpus_per_queue);
        if (cpus_weight(mask) < cpus_per_queue)
            nvme_add_cpus(&mask, unassigned_cpus,
                cpumask_of_node(
                    nvme_find_closest_node(
                        cpu_to_node(cpu))),
                nvmeq, cpus_per_queue);
        if (cpus_weight(mask) < cpus_per_queue)
            nvme_add_cpus(&mask, unassigned_cpus,
                unassigned_cpus,
                nvmeq, cpus_per_queue);

        WARN(cpumask_weight(nvmeq->cpu_mask) != cpus_per_queue,
            "nvme%d qid:%d mis-matched queue-to-cpu assignment\n",
            dev->instance, i);

        cpumask_andnot(unassigned_cpus, unassigned_cpus,
                        nvmeq->cpu_mask);
        cpu = cpumask_next(cpu, unassigned_cpus);
        if (remainder && !--remainder)
            cpus_per_queue++;
        unlock_nvmeq(nvmeq);
    }
    WARN(cpumask_weight(unassigned_cpus), "nvme%d unassigned online cpus\n",
                                dev->instance);
    i = 0;
    cpumask_andnot(unassigned_cpus, cpu_possible_mask, cpu_online_mask);
    for_each_cpu(cpu, unassigned_cpus)
        *per_cpu_ptr(dev->io_queue, cpu) = (i++ % queues) + 1;
    free_cpumask_var(unassigned_cpus);

    dev_dbg(NUB_KDEV(dev), "End.\n");
}

static int set_queue_count(struct nub_dev *dev, int count)
{
    int status;
    u32 result;
    u32 q_count = (count - 1) | ((count - 1) << 16);

    status = nvme_set_features(dev, NVME_FEAT_NUM_QUEUES, q_count, 0,
                                &result);
    dev_dbg(NUB_KDEV(dev), "Set_features return: status %x (result 0x%x)\n",
                status, result);

    if (status < 0)
        return status;
    if (status > 0) {
        dev_err(NUB_KDEV(dev), "Could not set queue count (%d)\n",
                                    status);
        return -EBUSY;
    }
    return min(result & 0xffff, result >> 16) + 1;
}

static size_t db_bar_size(struct nub_dev *dev, unsigned nr_io_queues)
{
    return 4096 + ((nr_io_queues + 1) * 8 * dev->db_stride);
}

static int nvme_setup_io_queues(struct nub_dev *dev)
{
    int result, nr_io_queues, size;
    int max_qdepth;
    size_t queue_space;
    size_t prp_pool_size;
    size_t num_bitmapped_pages, bitmap_size;

    dev_dbg(NUB_KDEV(dev), "Start.\n");
    /*
     * The user space driver needs to return back a value that is feasible
     * when considering the CMB given, as we'll try to use the CMB to supply
     * queue space and PRP lists for each queue pair.
     */
    nr_io_queues = num_possible_cpus();
    result = set_queue_count(dev, nr_io_queues);
    if (result < 0)
        return result;
    if (result < nr_io_queues)
        nr_io_queues = result;

    dev_dbg(NUB_KDEV(dev), "IO queue count: %d\n", nr_io_queues);
    size = db_bar_size(dev, nr_io_queues);
    if (size > dev->bar_size) {
        return -ENOMEM;
    }
    /*
     * Should investigate if there's a performance win from allocating
     * more queues than interrupt vectors; it might allow the submission
     * path to scale better, even if the receive path is limited by the
     * number of interrupts.
     */
    dev->max_qid = nr_io_queues;

    /* Update PRP list pool from CMB according to number of IO queues prescribed. */
    max_qdepth = max(dev->q_depth, DEFAULT_ADMIN_QDEPTH);
    queue_space = (nr_io_queues + 1) * (CQ_SIZE(max_qdepth) + SQ_SIZE(max_qdepth));
    queue_space = (queue_space + PAGE_SIZE) & ~(PAGE_SIZE-1);   // round up to next page size
    prp_pool_size = dev->cmb_size - queue_space;

    num_bitmapped_pages = prp_pool_size / PAGE_SIZE;
    bitmap_size = DIV_ROUND_UP(num_bitmapped_pages, sizeof(long)*8) * sizeof(long); // in bytes

    dev_dbg(NUB_KDEV(dev), "IO queue space: 0x%zx\n", queue_space);
    dev_dbg(NUB_KDEV(dev), "prp pool size: 0x%zx\n", prp_pool_size);
    if (dev->free_prpl_bitmap)
    {
        void* tmp = dev->free_prpl_bitmap;
        dev->free_prpl_bitmap = NULL;
        kfree(tmp);
    }
    dev->prpl_dma_addr = (uintptr_t) dev->cmb_pbase + queue_space;
    dev->prpl = dev->cmb_vbase + queue_space;
    dev->free_prpl_bitmap = kzalloc(bitmap_size, GFP_KERNEL);
    if (!dev->free_prpl_bitmap) {
        return -ENOMEM;
    }
    dev->prp_pool_size = num_bitmapped_pages;
    dev_dbg(NUB_KDEV(dev), "prp pool: 0x%p (DMA:0x%p)\n", dev->prpl, (void*)dev->prpl_dma_addr);

    /* Free previously allocated queues that are no longer usable */
    nvme_free_queues(dev, nr_io_queues + 1);
    dev_dbg(NUB_KDEV(dev), "Assigning IO queues...\n");
    nvme_assign_io_queues(dev);

    dev_dbg(NUB_KDEV(dev), "End.\n");
    return 0;
}

/*
 * Return: error value if an error occurred setting up the queues or calling
 * Identify Device.  0 if these succeeded, even if adding some of the
 * namespaces failed.  At the moment, these failures are silent.  TBD which
 * failures should be reported.
 */
static int nvme_dev_add(struct nub_dev *dev)
{
    int res;
    unsigned nn, i;
    struct nvme_ns *ns;
    struct nvme_id_ctrl *ctrl;
    struct nvme_id_ns *id_ns;
    void *mem;
    dma_addr_t dma_addr;
    int shift = NVME_CAP_MPSMIN(readq(&dev->bar->cap)) + 12;

    dev_dbg(NUB_KDEV(dev), "Start.\n");
    mem = nvme_alloc_cmd_buffer(dev, 8192, &dma_addr);
    if (!mem)
        return -ENOMEM;

    res = nvme_identify(dev, 0, 1, dma_addr);
    if (res) {
        dev_err(NUB_KDEV(dev), "Identify Controller failed (%d)\n", res);
        res = -EIO;
        goto out;
    }

    ctrl = mem;
    nn = le32_to_cpup(&ctrl->nn);
    dev->max_nsid = nn;
    dev->oncs = le16_to_cpup(&ctrl->oncs);
    dev->abort_limit = ctrl->acl + 1;
    dev->vwc = ctrl->vwc;
    memcpy(dev->serial, ctrl->sn, sizeof(ctrl->sn));
    memcpy(dev->model, ctrl->mn, sizeof(ctrl->mn));
    memcpy(dev->firmware_rev, ctrl->fr, sizeof(ctrl->fr));

    if (ctrl->mdts)
        dev->max_hw_sectors = 1 << (ctrl->mdts + shift - 9);
    else
        dev->max_hw_sectors = 4096; /* default to 2MB max xfer size */

    dev_dbg(NUB_KDEV(dev), "mdts:%d max_sectors:x%x\n", ctrl->mdts, dev->max_hw_sectors);

    id_ns = mem;
    for (i = 1; i <= nn; i++) {
        res = nvme_identify(dev, i, 0, dma_addr);
        dev_dbg(NUB_KDEV(dev), "NS identify for NSID %d: 0x%x\n", i, res);
        if (res)
            continue;

        if (id_ns->ncap == 0)
            continue;

        res = nvme_get_features(dev, NVME_FEAT_LBA_RANGE, i,
                            dma_addr + 4096, NULL);
        if (res)
            memset(mem + 4096, 0, 4096);

        ns = nvme_alloc_ns(dev, i, mem, mem + 4096);
        if (ns)
            list_add_tail(&ns->list, &dev->namespaces);
    }
    list_for_each_entry(ns, &dev->namespaces, list)
        add_disk(ns->disk);
    res = 0;

 out:
    nvme_free_cmd_buffer(dev, 8192, mem, dma_addr);
    dev_dbg(NUB_KDEV(dev), "End (returning 0x%x)\n", res);
    return res;
}

static int nvme_ns_update(struct nub_dev *dev)
{
    int res=0;
    unsigned i;

    dev_dbg(NUB_KDEV(dev), "Start.\n");

    for (i = 1; i <= dev->max_nsid; i++) {
        nubd_update_disk(dev, i);
        dev_dbg(NUB_KDEV(dev), "NS/disk update for NSID %d: 0x%x\n", i, res);
    }

    dev_dbg(NUB_KDEV(dev), "End (returning 0x%x)\n", res);
    return res;
}

static int nubd_update_disk(struct nub_dev *dev, int nsid)
{
    int res = 0;
    struct nvme_id_ns *ident;
    struct nvme_ns *ns;
    int is_found = 0;
    dma_addr_t dma_addr;
    void *mem;
    uint64_t disk_size=0;

    dev_dbg(NUB_KDEV(dev), "Start.\n");
    mem = nvme_alloc_cmd_buffer(dev, 8192, &dma_addr);
    if (!mem)
        return -ENOMEM;
    ident = mem;

    /* query state in controller */
    res = nvme_identify(dev, nsid, 0, dma_addr);
    dev_dbg(NUB_KDEV(dev), "NS identify for NSID %d: 0x%x\n", nsid, res);

    /* No change on error - return it up the stack for callers to handle. */
    if (res)
        goto free_ident_buf;

    /* query state in dev struct */
    list_for_each_entry(ns, &dev->namespaces, list)
    {
        if(ns->ns_id == nsid)
        {
            is_found = 1;

            /* ns now corresponds to the namespace in question. */
            break;
        }
    }
    dev_dbg(NUB_KDEV(dev), "Is found? %d(nsid %d) ns size: 0x%llx\n", is_found, nsid, ident->ncap);

    /* If in struct but gone now, remove it. */
    if (is_found && !ident->ncap)
    {
        dev_dbg(NUB_KDEV(dev), "Removing ns %d\n", nsid);
        /* Remove block device from system */
        if (ns->disk->flags & GENHD_FL_UP)
            del_gendisk(ns->disk);
        if (!blk_queue_dying(ns->queue))
            blk_cleanup_queue(ns->queue);

        /* Remove gendisk and namespace items. */
        list_del(&ns->list);
        put_disk(ns->disk);
        kfree(ns);
    }
    /* if missing in struct and now there, add it. */
    else if (!is_found && ident->ncap)
    {
        dev_dbg(NUB_KDEV(dev), "Adding ns %d\n", nsid);
        res = nvme_get_features(dev, NVME_FEAT_LBA_RANGE, nsid,
                            dma_addr + 4096, NULL);
        if (res)
            memset(mem + 4096, 0, 4096);

        res=0;
        ns = nvme_alloc_ns(dev, nsid, mem, mem + 4096);
        if (ns)
        {
            list_add_tail(&ns->list, &dev->namespaces);
            add_disk(ns->disk);
            dev_dbg(NUB_KDEV(dev), "Added disk %d\n", nsid);
        }
    }
    /* if in struct and in controller, update capacity and LBA range */
    else if (is_found && ident->ncap)
    {
        disk_size = le64_to_cpup(&ident->ncap) << (ns->lba_shift - 9);
        set_capacity(ns->disk, disk_size);
    }
    /* And if it is not found and no diskSize, then nothing to add, so nothing done. */

free_ident_buf:
    nvme_free_cmd_buffer(dev, 8192, mem, dma_addr);
    dev_dbg(NUB_KDEV(dev), "End. (returning 0x%x)\n", res);
    return res;
}

struct nvme_delq_ctx {
    struct task_struct *waiter;
    struct kthread_worker *worker;
    atomic_t refcount;
};

static void nvme_wait_dq(struct nvme_delq_ctx *dq, struct nub_dev *dev)
{
    dq->waiter = current;
    mb();

    for (;;) {
        set_current_state(TASK_KILLABLE);
        if (!atomic_read(&dq->refcount))
            break;
        if (!schedule_timeout(ADMIN_TIMEOUT) ||
                    fatal_signal_pending(current)) {
            set_current_state(TASK_RUNNING);

            nvme_disable_ctrl(dev, readq(&dev->bar->cap));
            nvme_disable_queue(dev, 0);

            send_sig(SIGKILL, dq->worker->task, 1);
            flush_kthread_worker(dq->worker);
            return;
        }
    }
    set_current_state(TASK_RUNNING);
}

static void nvme_put_dq(struct nvme_delq_ctx *dq)
{
    atomic_dec(&dq->refcount);
    if (dq->waiter)
        wake_up_process(dq->waiter);
}

static struct nvme_delq_ctx *nvme_get_dq(struct nvme_delq_ctx *dq)
{
    atomic_inc(&dq->refcount);
    return dq;
}

static void nvme_del_queue_end(struct nvme_queue *nvmeq)
{
    struct nvme_delq_ctx *dq = nvmeq->cmdinfo.ctx;

    nvme_clear_queue(nvmeq);
    nvme_put_dq(dq);
}

static int adapter_async_del_queue(struct nvme_queue *nvmeq, u8 opcode,
                        kthread_work_func_t fn)
{
    struct nvme_command c;

    memset(&c, 0, sizeof(c));
    c.delete_queue.opcode = opcode;
    c.delete_queue.qid = cpu_to_le16(nvmeq->qid);

    init_kthread_work(&nvmeq->cmdinfo.work, fn);
    return nvme_submit_admin_cmd_async(nvmeq->dev, &c, &nvmeq->cmdinfo);
}

static void nvme_del_cq_work_handler(struct kthread_work *work)
{
    struct nvme_queue *nvmeq = container_of(work, struct nvme_queue,
                            cmdinfo.work);
    nvme_del_queue_end(nvmeq);
}

static int nvme_delete_cq(struct nvme_queue *nvmeq)
{
    return adapter_async_del_queue(nvmeq, nvme_admin_delete_cq,
                        nvme_del_cq_work_handler);
}

static void nvme_del_sq_work_handler(struct kthread_work *work)
{
    struct nvme_queue *nvmeq = container_of(work, struct nvme_queue,
                            cmdinfo.work);
    int status = nvmeq->cmdinfo.status;

    if (!status)
        status = nvme_delete_cq(nvmeq);
    if (status)
        nvme_del_queue_end(nvmeq);
}

static int nvme_delete_sq(struct nvme_queue *nvmeq)
{
    return adapter_async_del_queue(nvmeq, nvme_admin_delete_sq,
                        nvme_del_sq_work_handler);
}

static void nvme_del_queue_start(struct kthread_work *work)
{
    struct nvme_queue *nvmeq = container_of(work, struct nvme_queue,
                            cmdinfo.work);
    allow_signal(SIGKILL);
    if (nvme_delete_sq(nvmeq))
        nvme_del_queue_end(nvmeq);
}

static void nvme_disable_io_queues(struct nub_dev *dev)
{
    int i;
    DEFINE_KTHREAD_WORKER_ONSTACK(worker);
    struct nvme_delq_ctx dq;
    struct task_struct *kworker_task = kthread_run(kthread_worker_fn,
                    &worker, "nvme%d", dev->instance);

    if (IS_ERR(kworker_task)) {
        dev_err(NUB_KDEV(dev), "Failed to create queue del task\n");
        for (i = dev->queue_count - 1; i > 0; i--)
            nvme_disable_queue(dev, i);
        return;
    }

    dq.waiter = NULL;
    atomic_set(&dq.refcount, 0);
    dq.worker = &worker;
    for (i = dev->queue_count - 1; i > 0; i--) {
        struct nvme_queue *nvmeq = raw_nvmeq(dev, i);

        if (nvme_suspend_queue(nvmeq))
            continue;
        nvmeq->cmdinfo.ctx = nvme_get_dq(&dq);
        nvmeq->cmdinfo.worker = dq.worker;
        init_kthread_work(&nvmeq->cmdinfo.work, nvme_del_queue_start);
        queue_kthread_work(dq.worker, &nvmeq->cmdinfo.work);
    }
    nvme_wait_dq(&dq, dev);
    kthread_stop(kworker_task);
}

/*
* Remove the node from the device list and check
* for whether or not we need to stop the nvme_thread.
*/
static void nvme_dev_list_remove(struct nub_dev *dev)
{
    struct task_struct *tmp = NULL;

    spin_lock(&dev_list_lock);
    list_del_init(&dev->node);
    if (list_empty(&dev_list) && !IS_ERR_OR_NULL(nvme_thread)) {
        tmp = nvme_thread;
        nvme_thread = NULL;
    }
    spin_unlock(&dev_list_lock);

    if (tmp)
    {
        dev_warn(NUB_KDEV(dev), "%s: Shutting down nvme kthread.\n", __FUNCTION__);
        kthread_stop(tmp);
    }
}

static void nvme_dev_shutdown(struct nub_dev *dev)
{
    int i;

    dev->initialized = 0;
    nvme_dev_list_remove(dev);

    if (!dev->bar) {
        for (i = dev->queue_count - 1; i >= 0; i--) {
            struct nvme_queue *nvmeq = raw_nvmeq(dev, i);
            nvme_suspend_queue(nvmeq);
            nvme_clear_queue(nvmeq);
        }
    } else {
        nvme_disable_io_queues(dev);
        nvme_disable_queue(dev, 0);
    }
}

static void nvme_dev_remove(struct nub_dev *dev)
{
    struct nvme_ns *ns;

    list_for_each_entry(ns, &dev->namespaces, list) {
        if (ns->disk->flags & GENHD_FL_UP)
            del_gendisk(ns->disk);
        if (!blk_queue_dying(ns->queue))
            blk_cleanup_queue(ns->queue);
    }
}

static void nvme_free_namespaces(struct nub_dev *dev)
{
    struct nvme_ns *ns, *next;

    list_for_each_entry_safe(ns, next, &dev->namespaces, list) {
        list_del(&ns->list);
        put_disk(ns->disk);
        kfree(ns);
    }
}

static void nvme_free_dev(struct kref *kref)
{
    struct nub_dev *dev = container_of(kref, struct nub_dev, kref);

    nvme_free_namespaces(dev);
    free_percpu(dev->io_queue);
    kfree(dev->queues);
    kfree(dev);
}

static int nubd_ctrl_open(struct inode *inode, struct file *f)
{
    struct nub_dev *dev = container_of(f->private_data, struct nub_dev,
                                miscdev);
    kref_get(&dev->kref);
    f->private_data = dev;
    return 0;
}

static int nubd_ctrl_release(struct inode *inode, struct file *f)
{
    struct nub_dev *dev = f->private_data;
    kref_put(&dev->kref, nvme_free_dev);
    return 0;
}

/* Returns an allocated array of addresses */
static u64* nubd_list_setup( u64 pptr1, u64 pptr2, size_t io_len, size_t offset)
{
    u64*    buf_list;
    u64*    adr_list;
    u64     page0_offset;
    int     n_prp_pages;
    int     xfer_pg0;
    size_t  list_size;

    page0_offset = (pptr1 + offset) & (PAGE_SIZE-1);
    n_prp_pages = (io_len + (pptr1 & (PAGE_SIZE-1)) + (PAGE_SIZE - 1)) / PAGE_SIZE;
    list_size = n_prp_pages * sizeof(void*);
    buf_list = kmalloc( list_size, GFP_KERNEL);
    if (!buf_list)
        return NULL;

    /* Finds first list entry. */
    xfer_pg0 = ((pptr1 & (PAGE_SIZE-1)) + offset) / PAGE_SIZE;
    switch (xfer_pg0) {
    case (0):   /* No offset */
        buf_list[0] = pptr1 + offset;
        if (n_prp_pages == 2)
        {
            buf_list[1] = pptr2;
        }
        else if (n_prp_pages > 2)
        {
            if (copy_from_user(&buf_list[1], (void*)pptr2, list_size-sizeof(void*)))
            {
                pr_warn("%s(%d): Failed copy from user space.(src: %p, dst: %p, len: %ld)\n",
                         __FUNCTION__, __LINE__, &buf_list[1], (void*)pptr2, list_size-sizeof(void*));
                kfree(buf_list);
                return NULL;
            }
        }
        break;
    case (1):
        if (n_prp_pages == 2)
        {
            buf_list[0] = pptr2 + page0_offset;
        }
        else if (n_prp_pages > 2)
        {
            if (copy_from_user(&buf_list[0], (void*)pptr2, list_size))
            {
                pr_warn("%s(%d): Failed copy from user space.(src: %p, dst: %p, len: %ld)\n",
                         __FUNCTION__, __LINE__, &buf_list[0], (void*)pptr2, list_size);
                kfree(buf_list);
                return NULL;
            }
            buf_list[0]+=page0_offset;
        }
        else
        {   /* Error */
            pr_warn("%s(%d): Ambiguous number of pages: (pptr1: %p, io_len: %ld, offset: %zd)\n",
                     __FUNCTION__, __LINE__, (void*)pptr1, io_len, offset);
            kfree(buf_list);
            return NULL;
        }
        break;
    default:
        adr_list = (u64*) pptr2;
        if (copy_from_user(&buf_list[0], &adr_list[xfer_pg0-1], list_size))
        {
            pr_warn("%s(%d): Failed copy from user space.(src: %p, dst: %p, len: %ld)\n",
                     __FUNCTION__, __LINE__, &buf_list[0], &adr_list[xfer_pg0-1], list_size);
            kfree(buf_list);
            return NULL;
        }
        buf_list[0]+=page0_offset;
        break;
    }
    return buf_list;
}

static int nubd_copy_data(struct nub_dev *dev,
                    struct nubd_rw_data_copy_req __user *ureq,
                    bool isWrite)
{
    struct nubd_rw_data_copy_req    req;
    size_t  remaining_bytes;
    size_t  copy_size;
    NubdSglDesc* server_sgl;
    NubdSglDesc* server_sge;
    u64     server_buf_ptr;
    void*   server_buf_vptr = NULL;
    u64*    client_buf_list = NULL;
    void*   client_buf_ptr;
    void*   client_buf_vptr;
    int     client_pg_cnt;
    u16     t10dif_crc = 0;
    u16     t10_app_tag;
    u32     t10_ref_tag;
    int     retval;
    size_t  num_clnt_bytes;
    size_t  num_svr_bytes;
    size_t  num_t10dif_bytes = 0;
    struct page*    client_pg = NULL;
    u16     pi_flags;
    struct t10_dif_tuple* pi_ptr;
    struct io_region_entry* region;

    dev_dbg(NUB_KDEV(dev), "Start.\n");
    if (copy_from_user(&req, ureq, sizeof(req)))
        return -EFAULT;

    pi_flags = req.pi_flags;
    t10_app_tag = req.pi_apptag;
    t10_ref_tag = req.pi_reftag;

    /* Sanity check provided parameters */
    if ((dev->max_hw_sectors && ((req.xfer_len / 512) > dev->max_hw_sectors))
        || !req.xfer_len
        || ((req.xfer_len < 512) && req.pi_flags)
        || !req.srv_sgl.addr
        || req.srv_sgl.flags.zero
        || !req.prp1)
        return -EINVAL;

    remaining_bytes = req.xfer_len;

    dev_dbg(NUB_KDEV(dev), "Copy REQ: sgl_nents: %zd, bytes x%zx, PI:x%x.\n",
            req.srv_sgl_nents, req.xfer_len, req.pi_flags);
    /*
     * Client buffer may not be aligned, so we do some
     * splitting and potentially make 2 passes to copy a page.
     */
    client_pg_cnt = 0;

    /* Server list setup. */
    copy_size =  req.srv_sgl_nents * sizeof(NubdSglDesc);
    server_sgl = (NubdSglDesc*) kmalloc(copy_size, GFP_KERNEL);
    if ( !server_sgl )
    {
        retval = -EFAULT;
        goto buf_list_free;
    }

    if (copy_from_user(server_sgl, (void*)req.srv_sgl.addr, copy_size))
    {
        retval = -EFAULT;
        goto buf_list_free;
    }

    server_sge = server_sgl;

    dev_dbg(NUB_KDEV(dev), "Setting up client list... %llx, %llx, %x, %x\n",
               req.prp1, req.prp2, req.prp_data_size, req.offset);
    /* Client list setup */
    client_buf_list = nubd_list_setup(req.prp1, req.prp2, req.prp_data_size, req.offset);
    if( !client_buf_list )
    {
        retval = -EFAULT;
        goto buf_list_free;
    }
    client_buf_ptr = (void*) client_buf_list[client_pg_cnt];

    server_buf_ptr = server_sge->addr;
    num_svr_bytes = server_sge->length;
    if ( pi_flags )
    {
        /*
         * If we have PI on the server for this transfer, we need
         * to adjust the byte count upward to account for the
         * extra 8 bytes PI for every 512B data.
         */
        num_svr_bytes = (520 * num_svr_bytes) / 512;
        num_t10dif_bytes = 512;
    }

    dev_dbg(NUB_KDEV(dev), "looking for 0x%llx.\n", server_buf_ptr);
    for(region=io_region_list; region; region=region->next)
    {
        dev_dbg(NUB_KDEV(dev), "region 0x%llx:0x%zx at %p.\n",
            region->p_addr, region->len, region->v_addr);

        if (    (region->p_addr <= server_buf_ptr)
            &&  (server_buf_ptr < (region->p_addr + region->len)))
            break;
    }
    if (!region || ((server_buf_ptr + num_svr_bytes) > (region->p_addr + region->len)))
    {
        dev_err(NUB_KDEV(dev), "Invalid copy req address range: 0x%llx:0x%zx. (PI:x%x)\n",
            server_buf_ptr, num_svr_bytes, pi_flags);
        retval = -EFAULT;
        goto buf_list_free;
    }
    server_buf_vptr = region->v_addr + (server_buf_ptr - region->p_addr);

    dev_dbg(NUB_KDEV(dev), "0x%zx svr bytes at %llx, mapped to %p.\n",
            num_svr_bytes, server_buf_ptr, server_buf_vptr);

    while (remaining_bytes)
    {
        dev_dbg(NUB_KDEV(dev), "bytes left: 0x%zx. sPtr %llx, cPtr %p.\n",
                remaining_bytes, server_buf_ptr, client_buf_ptr);
        if(client_pg_cnt + 1 == PAGE_SIZE / 8)
        {
            dev_warn(NUB_KDEV(dev), "Data copy request too large!\n");
            retval = -EINVAL;
            goto buf_list_free;
        }

        if( !server_buf_ptr  )
        {
            dev_warn(NUB_KDEV(dev), "Bad server address.\n");
            retval = -EINVAL;
            goto buf_list_free;
        }

        if( !client_buf_ptr  )
        {
            dev_warn(NUB_KDEV(dev), "Bad client address.\n");
            retval = -EINVAL;
            goto buf_list_free;
        }

        num_clnt_bytes = BYTES_LEFT_IN_PAGE(client_buf_ptr);

        /* Determine copy size for this iternation */
        copy_size = (num_clnt_bytes < num_svr_bytes)
                    ? min(num_clnt_bytes, remaining_bytes)
                    : min(num_svr_bytes, remaining_bytes);

        /* if transaction involves PI, limit to PI data block size. */
        if ( (pi_flags & NUBD_PI_HAS_PI) && copy_size > num_t10dif_bytes )
        {
            copy_size = num_t10dif_bytes;
        }

        client_pg = pfn_to_page( (u64)client_buf_ptr >> PAGE_SHIFT);
        client_buf_vptr = kmap(client_pg);
        if (!client_buf_vptr)
        {
            dev_warn(NUB_KDEV(dev), "kmap failed (%p)!\n", (void*)client_buf_list[client_pg_cnt]);
            retval = -EINVAL;
            goto buf_list_free;
        }
        client_buf_vptr += (u64)client_buf_ptr & (PAGE_SIZE-1);
        dev_dbg(NUB_KDEV(dev), "copy size: 0x%zx. clnt vaddr %p, srv vaddr %p.\n",
                copy_size, client_buf_vptr, server_buf_vptr);

        /* Copy the data, with t10 PI copy as needed. */
        if (isWrite)
        {
            if ( pi_flags & NUBD_PI_HAS_PI )
            {
                t10dif_crc = crc16_t10dif_copy_by4( t10dif_crc, server_buf_vptr, client_buf_vptr, copy_size );
            }
            else
            {
                memcpy(server_buf_vptr, client_buf_vptr, copy_size);
            }
        }
        else
        {
            if ( pi_flags & NUBD_PI_HAS_PI )
            {
                t10dif_crc = crc16_t10dif_copy_by4( t10dif_crc, client_buf_vptr, server_buf_vptr, copy_size );
            }
            else
            {
                memcpy(client_buf_vptr, server_buf_vptr, copy_size);
            }
        }

        kunmap(client_pg);

        /* Client copied before PI adjust since no PI in client. */
        client_buf_ptr += copy_size;
        client_buf_vptr += copy_size;
        if ( IS_ALIGNED((u64)client_buf_ptr, PAGE_SIZE) )
        {
            client_pg_cnt++;
            client_buf_ptr = (void*)client_buf_list[client_pg_cnt];
        }

        remaining_bytes -= copy_size;

        server_buf_ptr += copy_size;
        server_buf_vptr += copy_size;
        num_svr_bytes -= copy_size;

        /* Process PI as needed, updating pointer as appropriate. */
        if ( pi_flags & NUBD_PI_HAS_PI )
        {
            num_t10dif_bytes -= copy_size;
            if ( !num_t10dif_bytes )
            {
                pi_ptr = server_buf_vptr;
                dev_dbg(NUB_KDEV(dev), "PI %d: vptr %p, g %x: at %x: rt %x.\n", isWrite,
                        pi_ptr, pi_ptr->guard_tag, pi_ptr->app_tag, pi_ptr->ref_tag);
                if (isWrite)
                {
                    pi_ptr->guard_tag = cpu_to_be16( t10dif_crc );
                    pi_ptr->app_tag = cpu_to_be16( t10_app_tag );
                    pi_ptr->ref_tag = cpu_to_be32( t10_ref_tag );
                }
                else
                {
                    if (pi_ptr->guard_tag !=  cpu_to_be16(t10dif_crc))
                    {
                        dev_warn(NUB_KDEV(dev), "PI miscompare: calc: 0x%x, PI: 0x%x\n", t10dif_crc, pi_ptr->guard_tag );
                        retval = -EIO;
                        goto buf_list_free;
                    }
                    if ( (pi_flags & NUBD_PI_CHECK_APPTAG)
                        && (pi_ptr->app_tag != cpu_to_be16(t10_app_tag)))
                    {
                        dev_warn(NUB_KDEV(dev), "PI bad app tag: expected: 0x%x, observed: 0x%x\n", 0, pi_ptr->app_tag );
                        retval = -EIO;
                        goto buf_list_free;
                    }
                    if ( (pi_flags & NUBD_PI_CHECK_REFTAG)
                        && (pi_ptr->ref_tag !=  cpu_to_be32(t10_ref_tag)))
                    {
                        dev_warn(NUB_KDEV(dev), "PI bad ref tag: expected: 0x%x, observed: 0x%x\n", 0, pi_ptr->ref_tag );
                        retval = -EIO;
                        goto buf_list_free;
                    }
                }

                t10_ref_tag++;
                t10dif_crc = 0;

                server_buf_ptr += 8;
                server_buf_vptr += 8;
                num_svr_bytes -= 8;
                num_t10dif_bytes = 512;
            }
        }

        if ( !num_svr_bytes && remaining_bytes )
        {
            /* Move to the next SGE entry... */
            server_sge++;
            server_buf_ptr = server_sge->addr;
            num_svr_bytes = server_sge->length;
            if ( pi_flags & NUBD_PI_HAS_PI )
            {
                num_svr_bytes = (520 * num_svr_bytes) / 512;
            }

            for(region=io_region_list; region; region=region->next)
            {
                if (    (region->p_addr <= server_buf_ptr)
                    &&  (server_buf_ptr < (region->p_addr + region->len)))
                    break;
            }
            if (!region || ((server_buf_ptr + num_svr_bytes) > (region->p_addr + region->len)))
            {
                dev_err(NUB_KDEV(dev), "Invalid copy req address range: 0x%llx:0x%zxi (PI:x%x)\n",
                    server_buf_ptr, num_svr_bytes, pi_flags);
                retval = -EFAULT;
                goto buf_list_free;
            }
            server_buf_vptr = region->v_addr + (server_buf_ptr - region->p_addr);

            dev_dbg(NUB_KDEV(dev), "0x%zx svr bytes at %llx, mapped to %p.\n",
                num_svr_bytes, server_buf_ptr, server_buf_vptr);
        }
    }

    retval = 0;

buf_list_free:
    kfree(server_sgl);
    kfree(client_buf_list);
    dev_dbg(NUB_KDEV(dev), "End (returning 0x%x).\n", retval);
    return retval;
}

static long nubd_ctrl_ioctl(struct file *f, unsigned int cmd, unsigned long arg)
{
    int result = 0;
    int qid;
    int nsid;
    struct nub_dev *dev = f->private_data;
    struct nubd_io_mem_range new_iomem;
    struct io_region_entry* new_region;

    /* Only let Admin CQ tail updates through if the device is not fully initialized.
     * Admin CQ updates are necessary to complete device init.
     */
    if((cmd!=NUBD_IOCTL_CQ_TAIL_UPDATE)
        && (!dev->initialized))
        return -ENOTTY;

    switch (cmd) {
    case NUBD_IOCTL_START_CTRL:
        /* Prevent concurrent namespace changes. */
        if (down_timeout(&ns_update_lock, NAMESPACE_LOCK_TIMEOUT * HZ))
        {
            dev_err(NUB_KDEV(dev), "Timeout from concurrent namespace modifications\n");
            return -EBUSY;
        }

        result = nvme_dev_add(dev);
        /* Failing to add disks, tear down the whole device. */
        if (result)
        {
            nubd_remove(dev);
        }

        up(&ns_update_lock);
        return result;
    case NUBD_IOCTL_RESCAN_DISKS:
        /* Prevent concurrent namespace changes. */
        if (down_timeout(&ns_update_lock, NAMESPACE_LOCK_TIMEOUT * HZ))
        {
            dev_err(NUB_KDEV(dev), "Timeout from concurrent namespace modifications\n");
            return -EBUSY;
        }

        result = nvme_ns_update(dev);

        up(&ns_update_lock);
        return result;
    case NUBD_IOCTL_CQ_TAIL_UPDATE:
        qid = (int) arg;
        if ((qid < 0) || (qid > dev->queue_count))
            return -EINVAL;
        /*
         * Running vdbench we can get a good boost in IOPs and a smaller but noticeable
         * one for bandwidth if we defer the completion queue processing to the client
         * context.  If the container needs a bit more CPU horsepower the clnt_completion
         * parameter can be cleared for the system to have the user space driver
         * incur that overhead instead.
         */
        if (clnt_completion)
        {
            queue_work( dev->queues[qid]->io_workqueue, &dev->queues[qid]->cq_update_work );
        }
        else
        {
            nvme_irq(0, dev->queues[qid]);
        }
        return 0;
    case NUBD_IOCTL_DISK_UPDATE:
        nsid = (int) arg;

        if ((nsid <= 0) || (nsid > dev->max_nsid))
            return -EINVAL;

        /* Prevent concurrent namespace changes. */
        if (down_timeout(&ns_update_lock, NAMESPACE_LOCK_TIMEOUT * HZ))
        {
            dev_err(NUB_KDEV(dev), "Timeout from concurrent namespace modifications\n");
            return -EBUSY;
        }

        result = nubd_update_disk(dev, nsid);

        up(&ns_update_lock);
        return result;
    case NUBD_IOCTL_RD_COMPLETE:
        result = nubd_copy_data(dev, (void __user *)arg, false);
        return result;
    case NUBD_IOCTL_WR_COMPLETE:
        result = nubd_copy_data(dev, (void __user *)arg, true);
        return result;
    case NUBD_IOCTL_ADD_IOMEM:
        if (copy_from_user( &new_iomem, (struct nubd_io_mem_range  __user*) arg,
                            sizeof(new_iomem)))
            return -EFAULT;

        /* would keeping the region list sorted help? */
        /* Do we need to check for overlapping/duplicate regions? */

        new_region = kzalloc(sizeof(struct io_region_entry), GFP_KERNEL);
        if (!new_region)
            return -ENOMEM;
        new_region->p_addr = new_iomem.p_addr;
        new_region->len = new_iomem.length;
        new_region->v_addr = ioremap_cache( new_iomem.p_addr,  new_iomem.length);
        new_region->next = io_region_list;

        if ( !new_region->v_addr )
        {
            dev_err(NUB_KDEV(dev), "Failed mapping iomem region (0x%llx:0x%zx)\n",
                    new_iomem.p_addr, new_iomem.length);
            kfree(new_region);
            return -EINVAL;
        }

        io_region_list = new_region;
        dev_dbg(NUB_KDEV(dev), "Mapped in region (0x%llx:0x%zx) at %p.\n",
            new_iomem.p_addr, new_iomem.length, new_region->v_addr);

        return result;
    default:
        dev_warn(NUB_KDEV(dev), "Bad ioctl request (0x%x)\n", cmd);
        return -ENOTTY;
    }
}

static const struct file_operations nubd_ctrl_fops = {
    .owner      = THIS_MODULE,
    .open       = nubd_ctrl_open,
    .release    = nubd_ctrl_release,
    .unlocked_ioctl = nubd_ctrl_ioctl,
    .compat_ioctl   = nubd_ctrl_ioctl,
};

static int nvme_dev_start(struct nub_dev *dev)
{
    int result;
    bool start_thread = false;
    u64 cap;

    /*
     * On a real device we would map in PCI memory-mapped reg space,
     * but here we just need to harvest the key values from the
     * pseudo-register space provided by the userland driver.
     */
    dev_dbg(NUB_KDEV(dev), "Getting DB  and queue parameters...\n");
    cap = readq(&dev->bar->cap);
    dev->q_depth = min_t(int, NVME_CAP_MQES(cap) + 1, NVME_Q_DEPTH);
    dev->db_stride = 1 << NVME_CAP_STRIDE(cap);
    dev->dbs = ((void __iomem *)dev->bar) + 0x1000;

    dev_dbg(NUB_KDEV(dev), "setup adminq...\n");
    result = nvme_configure_admin_queue(dev);
    if (result)
        return result;

    dev_dbg(NUB_KDEV(dev), "kernel thread startup...\n");
    spin_lock(&dev_list_lock);
    if (list_empty(&dev_list) && IS_ERR_OR_NULL(nvme_thread)) {
        start_thread = true;
        nvme_thread = NULL;
    }
    list_add(&dev->node, &dev_list);
    spin_unlock(&dev_list_lock);

    if (start_thread) {
        nvme_thread = kthread_run(nvme_kthread, NULL, "nubd");
        wake_up(&nvme_kthread_wait);
    } else
        wait_event_killable(nvme_kthread_wait, nvme_thread);

    if (IS_ERR_OR_NULL(nvme_thread)) {
        result = nvme_thread ? PTR_ERR(nvme_thread) : -EINTR;
        goto disable;
    }

    dev_dbg(NUB_KDEV(dev), "setup ioq...\n");
    result = nvme_setup_io_queues(dev);
    if (result && result != -EBUSY)
        goto disable;

    dev_dbg(NUB_KDEV(dev), "returning successful...\n");
    return result;

 disable:
    nvme_disable_queue(dev, 0);
    nvme_dev_list_remove(dev);
    return result;
}

static int nvme_remove_dead_ctrl(void *arg)
{
    struct nub_dev *dev = (struct nub_dev *)arg;

    kref_put(&dev->kref, nvme_free_dev);
    return 0;
}

static void nvme_remove_disks(struct work_struct *ws)
{
    struct nub_dev *dev = container_of(ws, struct nub_dev, reset_work);

    nvme_dev_remove(dev);
    nvme_free_queues(dev, 1);
}

static int nvme_dev_resume(struct nub_dev *dev)
{
    int ret;

    ret = nvme_dev_start(dev);
    if (ret && ret != -EBUSY)
        return ret;
    if (ret == -EBUSY) {
        spin_lock(&dev_list_lock);
        dev->reset_workfn = nvme_remove_disks;
        queue_work(nvme_workq, &dev->reset_work);
        spin_unlock(&dev_list_lock);
    }
    dev->initialized = 1;
    return 0;
}

static void nvme_reset_failed_dev(struct work_struct *ws)
{
    struct nub_dev *dev = container_of(ws, struct nub_dev, reset_work);

    nvme_dev_shutdown(dev);
    if (nvme_dev_resume(dev)) {
        dev_err(NUB_KDEV(dev), "Device failed to resume\n");
        kref_get(&dev->kref);
        if (IS_ERR(kthread_run(nvme_remove_dead_ctrl, dev, "nvme%d",
                            dev->instance))) {
            dev_err(NUB_KDEV(dev),
                "Failed to start controller remove task\n");
            kref_put(&dev->kref, nvme_free_dev);
        }
    }
}

static void nvme_reset_workfn(struct work_struct *work)
{
    struct nub_dev *dev = container_of(work, struct nub_dev, reset_work);
    dev->reset_workfn(work);
}


/**************************************/


/* Initializes the device structure.
 *
 * dev: Pointer to zeroed-out nub_dev structure.
 * usrconfig: pointer to struct from user with parameters of new device.
 *
 * This function is to set up the device structure, ready the admin queue for use
 * and have the memory spaces for the ioqueues figured out.
 *
 * We defer namespace discovery to another step to make sure the user space
 * part of the system is ready for it.
 */
static int nubd_ctrl_init(struct nub_dev* dev,
        struct nubd_ctrl_config* usrConfig)
{
    int result = -ENOMEM;
    u32 max_io_queues;

    dev_dbg(NUB_KDEV(dev), "Start.\n");
    /* Sanity check given base values and cmb size */
    if ( !(usrConfig->cmb_base)
        || (usrConfig->cmb_size < 1024*1024)
        || ((uintptr_t)usrConfig->cmb_base & (PAGE_SIZE-1))
        || !(usrConfig->reg_base))
        return -EINVAL;

    /*
     * We might be able to reduce this with input, but for now
     * just allocate for all possible cpus.  Should be relatively small
     * amount of "wasted" memory.
     */
    max_io_queues = num_possible_cpus();

    /* workqueues for IO CQ update notifications from user space. */
    dev->io_workqueues = kcalloc(max_io_queues, sizeof(*dev->io_workqueues),
                                GFP_KERNEL);
    if (!dev->io_workqueues)
        goto free;

    /* list of NVME queue pair structures (plus 1 for admin queue) */
    dev->queues = kcalloc(max_io_queues + 1, sizeof(void *),
                                GFP_KERNEL);
    if (!dev->queues)
        goto free;

    /* QIDs, per CPU. */
    dev->io_queue = alloc_percpu(unsigned short);
    if (!dev->io_queue)
        goto free;

    INIT_LIST_HEAD(&dev->namespaces);
    dev->reset_workfn = nvme_reset_failed_dev;
    INIT_WORK(&dev->reset_work, nvme_reset_workfn);

    /* Set up cmb for use. */
    dev->cmb_pbase = usrConfig->cmb_base;
    dev->cmb_unused_offset = 0;
    dev_dbg(NUB_KDEV(dev), "Mapping CMB (%p) of size 0x%zx\n", usrConfig->cmb_base, usrConfig->cmb_size);
    dev->cmb_vbase = ioremap_cache((phys_addr_t)usrConfig->cmb_base, usrConfig->cmb_size);
    dev->cmb_size = usrConfig->cmb_size;

    /* Map in pseudo-reg space provided by user-space. */
    dev_dbg(NUB_KDEV(dev), "Mapping pseudo regs (%p) of size 0x%zx\n", usrConfig->reg_base, usrConfig->reg_size);
    dev->bar = ioremap_cache((phys_addr_t)usrConfig->reg_base, usrConfig->reg_size);
    dev->bar_size = usrConfig->reg_size;

    if (!dev->bar || !dev->cmb_vbase)
        goto unmap;

    kref_init(&dev->kref);

    /*
     *  Note: Current naming convention gives us 26 possible controller devices,
     *      nubd-a through nubd-z.  However we only use one at this point.
     *  We need to create the device prior to calling nvme_dev_start() so it
     *  can use the controller ioctls for adminQ update notification.
     */
    dev_dbg(NUB_KDEV(dev), "Adding Ctrl device.\n");
    snprintf(dev->name, sizeof(dev->name), "nubd-%c", 'a' + dev->instance );
    dev->miscdev.minor = MISC_DYNAMIC_MINOR;
    dev->miscdev.name = dev->name;
    dev->miscdev.parent = nubd_mgr.this_device;
    dev->miscdev.fops = &nubd_ctrl_fops;
    result = misc_register(&dev->miscdev);
    if (result)
        goto remove;

    dev_dbg(NUB_KDEV(dev), "calling dev start...\n");
    result = nvme_dev_start(dev);
    if (result && result != -EBUSY) {
            goto free_pools;
    }

    dev->initialized = 1;
    dev_dbg(NUB_KDEV(dev), "End.\n");
    return 0;

 free_pools:
    nvme_free_queues(dev, 0);
 remove:
    nvme_dev_shutdown(dev);
 unmap:
    iounmap(dev->cmb_vbase);
    iounmap(dev->bar);
 free:
    free_percpu(dev->io_queue);
    kfree(dev->queues);
    kfree(dev->io_workqueues);
    kfree(dev);
    return result;
}

static void nubd_remove(struct nub_dev *dev)
{
    spin_lock(&dev_list_lock);
    list_del_init(&dev->node);
    spin_unlock(&dev_list_lock);

    flush_work(&dev->reset_work);
    misc_deregister(&dev->miscdev);
    nvme_dev_remove(dev);
    nvme_dev_shutdown(dev);
    nvme_free_queues(dev, 0);
    rcu_barrier();
    iounmap(dev->cmb_vbase);
    iounmap(dev->bar);
    kfree(dev->io_workqueues);
    ida_simple_remove(&nubd_ctrl_ida, dev->instance);
    kref_put(&dev->kref, nvme_free_dev);
}

/*
 * We only really support one controller now, but we tried to put in the
 * framework for being extensible if we ever found we needed it.
 */

static int nubd_mgr_create_ctrl(struct nubd_ctrl_config *ctrl_config)
{
    int result = 0;
    int new_dev_num;
    struct nub_dev* new_ctrl;

    pr_debug("Start.\n");
    /* create char-misc device.  Add to list. */
    new_ctrl = kzalloc( sizeof(struct nub_dev), GFP_KERNEL);
    if (!new_ctrl)
        return -ENOMEM;

    new_dev_num = ida_simple_get(&nubd_ctrl_ida, 0, max_ctrls, GFP_KERNEL);
    if (new_dev_num < 0)
    {
        kfree(new_ctrl);
        return new_dev_num;
    }

    new_ctrl->instance = new_dev_num;
    result = nubd_ctrl_init(new_ctrl, ctrl_config);

    pr_debug("controller init result: 0x%x\n", result);
    if (result)
    {
        ida_simple_remove(&nubd_ctrl_ida, new_dev_num);
        kfree(new_ctrl);
    }
    else
    {
        ctrl_config->handle = new_dev_num;
        nubd_ctrls[new_dev_num] = new_ctrl;
    }

    pr_debug("End.\n");
    return result;
}


static int nubd_mgr_destroy_ctrl(int nubd_ctrl_num)
{
    struct nub_dev *dev;

    /* Sanity check the given device handle. */
    if ( (nubd_ctrl_num >= max_ctrls)
        || (nubd_ctrl_num < 0)
        || !(nubd_ctrls[nubd_ctrl_num])
       )
    {
        return -EINVAL;
    }

    dev = nubd_ctrls[nubd_ctrl_num];
    nubd_remove(dev);

    return 0;
}
#endif


static long nubd_mgr_ioctl(struct file *f, unsigned int cmd, unsigned long arg)
{
#ifndef DISABLE_NUBD_DRIVER
    struct nubd_ctrl_config new_ctrl_cfg;
    int result = 0;

    switch (cmd) {
    case NUBD_IOCTL_CREATE_DEV:
        if (copy_from_user( &new_ctrl_cfg, (struct nubd_ctrl_config __user*) arg,
                            sizeof(new_ctrl_cfg)))
            return -EFAULT;

        result = nubd_mgr_create_ctrl(&new_ctrl_cfg);
        if (copy_to_user((struct nubd_ctrl_config __user*) arg, &new_ctrl_cfg,
                            sizeof(new_ctrl_cfg)))
            return -EFAULT;

        return result;
    case NUBD_IOCTL_DESTROY_DEV:
        return nubd_mgr_destroy_ctrl((int) arg);
    default:
        return -ENOTTY;
    }
#else
    return -EINVAL;
#endif
}

static const struct file_operations nubd_mgr_fops = {
    .owner      = THIS_MODULE,
    .unlocked_ioctl = nubd_mgr_ioctl,
    .compat_ioctl   = nubd_mgr_ioctl
};

static int __init nubd_init(void)
{
#ifndef DISABLE_NUBD_DRIVER
    int result;

    init_waitqueue_head(&nvme_kthread_wait);

    nvme_workq = create_singlethread_workqueue("nubd");
    if (!nvme_workq)
        return -ENOMEM;

    nubd_ctrls = kzalloc( max_ctrls * sizeof(struct miscdevice*), GFP_KERNEL);
    if (!nubd_ctrls)
    {
        result = -ENOMEM;
        goto remove_workq;
    }

    nubd_mgr.minor = MISC_DYNAMIC_MINOR;
    nubd_mgr.name = "nubd";
    nubd_mgr.fops = &nubd_mgr_fops;
    result = misc_register (&nubd_mgr);
    if (result)
        goto free_dev_array;

    result = register_blkdev(nubd_major, "nubd");
    if (result < 0)
    {
        goto unregister_mgr;
    }
    else if (result > 0)
        nubd_major = result;

    return 0;

unregister_mgr:
    misc_deregister(&nubd_mgr);
free_dev_array:
    kfree(nubd_ctrls);
remove_workq:
    destroy_workqueue(nvme_workq);
    return (result);
#else
    return 0;
#endif
}

static void __exit nubd_exit(void)
{
#ifndef DISABLE_NUBD_DRIVER
    int i;

    pr_debug("Start.\n");
    for (i=0; i< max_ctrls; ++i)
    {
        nubd_mgr_destroy_ctrl(i);
        pr_debug("Ctrl %d removed...\n", i);
    }

    unregister_blkdev(nubd_major, "nubd");
    nubd_major = 0;
    misc_deregister(&nubd_mgr);
    kfree(nubd_ctrls);
    destroy_workqueue(nvme_workq);
    BUG_ON(nvme_thread && !IS_ERR(nvme_thread));
    pr_debug("End.\n");
#endif
}

MODULE_AUTHOR("NetApp <noreply@netapp.com>");
MODULE_LICENSE("GPL");
MODULE_VERSION("0.9");
module_init(nubd_init);
module_exit(nubd_exit);
